/***************************************************************************
 *   Copyright (C) 2003 by Sbastien Laot                                 *
 *   slaout@linux62.org                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qdragobject.h>
#include <qdom.h>
#include <qpainter.h>
#include <qstyle.h>
#include <kstyle.h>
#include <qtooltip.h>
#include <qlistview.h>
#include <qcursor.h>
#include <qsimplerichtext.h>
#include <ktextedit.h>
#include <qpoint.h>
#include <qstringlist.h>
#include <kapplication.h>
#include <kglobalsettings.h>
#include <kopenwith.h>
#include <kservice.h>
#include <klocale.h>
#include <kglobalaccel.h>
#include <qdir.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <kfiledialog.h>
#include <kaboutdata.h>

#include <kpopupmenu.h>
#include <kiconloader.h>
#include <krun.h>

#include <qtoolbar.h>
#include <qclipboard.h>

#include <kmessagebox.h>
#include <qinputdialog.h>

#include <qlayout.h>

#include <stdlib.h>     // rand() function
#include <qdatetime.h>  // seed for rand()

#include "basket.h"
#include "item.h"
#include "itemdrag.h"
#include "itemfactory.h"
#include "itemedit.h"
#include "xmlwork.h"
#include "global.h"
#include "backgroundsmanager.h"
#include "settings.h"
#include "utilities.h"
#include "debugwindow.h"
#include "exporterdialog.h"
#include "config.h"


#include <iostream>

ItemSelection* ItemSelection::nextStacked()
{
	// First, search in the childs:
	if (firstChild)
		if (firstChild->item && firstChild->item->content())
			return firstChild;
		else
			return firstChild->nextStacked();

	// Then, in the next:
	if (next)
		if (next->item && next->item->content())
			return next;
		else
			return next->nextStacked();

	// And finally, in the parent:
	ItemSelection *node = parent;
	while (node)
		if (node->next)
			if (node->next->item && node->next->item->content())
				return node->next;
			else
				return node->next->nextStacked();
		else
			node = node->parent;

	// Not found:
	return 0;
}

ItemSelection* ItemSelection::firstStacked()
{
	if (!this)
		return 0;

	if (item && item->content())
		return this;
	else
		return nextStacked();
}

void ItemSelection::append(ItemSelection *node)
{
	if (!this || !node)
		return;

	if (firstChild) {
		ItemSelection *last = firstChild;
		while (last->next)
			last = last->next;
		last->next = node;
	} else
		firstChild = node;

	while (node) {
		node->parent = this;
		node = node->next;
	}
}

int ItemSelection::count()
{
	if (!this)
		return 0;

	int count = 0;
//	if (item && item->content())
//		count = 1;

	for (ItemSelection *node = this; node; node = node->next)
		if (node->item && node->item->content())
			++count;
		else
			count += node->firstChild->count();

	return count;
}

QValueList<Item*> ItemSelection::parentGroups()
{
	QValueList<Item*> groups;

	// For each item:
	for (ItemSelection *node = firstStacked(); node; node = node->nextStacked())
		// For each parent groups of the item:
		for (Item *item = node->item->parentItem(); item; item = item->parentItem())
			// Add it (if it was not already in the list):
			if (!item->isColumn() && !groups.contains(item))
				groups.append(item);

	return groups;
}


/** DecoratedBasket */

DecoratedBasket::DecoratedBasket(QWidget *parent, const QString &folderName, const char *name, WFlags fl)
 : QWidget(parent, name, fl)
{
	m_layout = new QVBoxLayout(this);
	m_search = new SearchBar(this);
	m_basket = new Basket(this, folderName);
	m_layout->addWidget(m_basket);
	setSearchBarPosition(Settings::searchOnTop());

	m_search->setShown(true);//m_basket->showSearchBar());
	m_basket->setFocus(); // To avoid the search bar have focus on load

	connect( m_search, SIGNAL(newSearch(const SearchData&)), m_basket, SLOT(newSearch(const SearchData&)) );

	connect( m_basket, SIGNAL(postMessage(const QString&)),      Global::mainContainer, SLOT(postStatusbarMessage(const QString&)) );
	connect( m_basket, SIGNAL(setStatusBarText(const QString&)), Global::mainContainer, SLOT(setStatusBarHint(const QString&))     );
	connect( m_basket, SIGNAL(resetStatusBarText()),             Global::mainContainer, SLOT(updateStatusBarHint())                );
}

DecoratedBasket::~DecoratedBasket()
{
}

void DecoratedBasket::setSearchBarPosition(bool onTop)
{
	m_layout->remove(m_search);
	if (onTop) {
		m_layout->insertWidget(0, m_search);
		setTabOrder(this/*(QWidget*)parent()*/, m_search);
		setTabOrder(m_search, m_basket);
		setTabOrder(m_basket, (QWidget*)parent());
	} else {
		m_layout->addWidget(m_search);
		setTabOrder(this/*(QWidget*)parent()*/, m_basket);
		setTabOrder(m_basket, m_search);
		setTabOrder(m_search, (QWidget*)parent());
	}
}

void DecoratedBasket::setSearchBarShown(bool show, bool switchFocus)
{
	m_basket->setShowSearchBar(true);//show);
//	m_basket->save();
	// In this order (m_basket and then m_search) because setShown(false)
	//  will call resetSearch() that will update actions, and then check the
	//  Ctrl+F action whereas it should be unchecked
	//  FIXME: It's very uggly all those things
	m_search->setShown(true);//show);
	if (show) {
		if (switchFocus)
			m_search->setEditFocus();
	} else if (m_search->hasEditFocus())
		m_basket->setFocus();
}

void DecoratedBasket::resetSearch()
{
	m_search->reset();
}












const int Basket::FRAME_DELAY = 50/*1500*/; // Delay between two animation "frames" in milliseconds

/*
 * Convenient function (defined in bsvitem.cpp !):
 */
void drawGradient( QPainter *p, const QColor &colorTop, const QColor & colorBottom,
				   int x, int y, int w, int h,
				   bool sunken, bool horz, bool flat  );

/*
 * Also defined in bsvitem.cpp):
 */
extern void substractRectOnAreas(const QRect &rectToSubstract, QValueList<QRect> &areas, bool andRemove = true);




#include <iostream>
void debugZone(int zone)
{
	QString s;
	switch (zone) {
		case Item::Handle:        s = "Handle";              break;
		case Item::Group:         s = "Group";               break;
		case Item::TagsArrow:     s = "TagsArrow";           break;
		case Item::Custom0:       s = "Custom0";             break;
		case Item::GroupExpander: s = "GroupExpander";       break;
		case Item::Content:       s = "Content";             break;
		case Item::Link:          s = "Link";                break;
		case Item::TopInsert:     s = "TopInsert";           break;
		case Item::TopGroup:      s = "TopGroup";            break;
		case Item::BottomInsert:  s = "BottomInsert";        break;
		case Item::BottomGroup:   s = "BottomGroup";         break;
		case Item::BottomColumn:  s = "BottomColumn";        break;
		case Item::None:          s = "None";                break;
		default:
			if (zone == Item::Emblem0)
				s = "Emblem0";
			else
				s = "Emblem0+" + QString::number(zone - Item::Emblem0);
			break;
	}
	std::cout << s << std::endl;
}

#define FOR_EACH_ITEM(itemVar) \
	for (Item *itemVar = firstItem(); itemVar; itemVar = itemVar->next())



void Basket::prependItemIn(Item *item, Item *in)
{
	if (!item)
		// No item to prepend:
		return;

	if (in) {
		// The normal case:
		preparePlug(item);

		Item *last = item->lastSibling();

		for (Item *i = item; i; i = i->next())
			i->setParentItem(in);
//		item->setPrev(0L);
		last->setNext(in->firstChild());

		if (in->firstChild())
			in->firstChild()->setPrev(last);

		in->setFirstChild(item);

		finishPlug();
	} else
		// Prepend it directly in the basket:
		appendItemBefore(item, firstItem());
}

void Basket::appendItemIn(Item *item, Item *in)
{
	if (!item)
		// No item to append:
		return;

	if (in) {
		// The normal case:
		preparePlug(item);

//		Item *last = item->lastSibling();
		Item *lastChild = in->lastChild();

		for (Item *i = item; i; i = i->next())
			i->setParentItem(in);
		item->setPrev(lastChild);
//		last->setNext(0L);

		if (!in->firstChild())
			in->setFirstChild(item);

		if (lastChild)
			lastChild->setNext(item);

		finishPlug();
	} else
		// Prepend it directly in the basket:
		appendItemAfter(item, lastItem());
}

void Basket::appendItemAfter(Item *item, Item *after, bool alreadyInBasket /* = false*/)
{
	if (!item)
		// No item to append:
		return;

	if (!after)
		// By default, insert after the last item:
		after = lastItem();

//	if (m_loaded && !item->isPrimary())
	if (m_loaded && after && !after->isFree() && !after->isColumn())
		for (Item *i = item; i; i = i->next())
			i->inheritTagsOf(after);

	if (!alreadyInBasket)
		preparePlug(item);

	Item *last = item->lastSibling();
	if (after) {
		// The normal case:
		for (Item *i = item; i; i = i->next())
			i->setParentItem(after->parentItem());
		item->setPrev(after);
		last->setNext(after->next());
		after->setNext(item);
		if (last->next())
			last->next()->setPrev(last);
	} else {
		// There is no item in the basket:
		for (Item *i = item; i; i = i->next())
			i->setParentItem(0);
		m_firstItem = item;
//		item->setPrev(0);
//		last->setNext(0);
	}

	if (!alreadyInBasket)
		finishPlug();
}

void Basket::appendItemBefore(Item *item, Item *before)
{
	if (!item)
		// No item to append:
		return;

	if (!before)
		// By default, insert before the first item:
		before = firstItem();

//	if (m_loaded && !item->isPrimary())
	if (m_loaded && before && !before->isFree() && !before->isColumn())
		for (Item *i = item; i; i = i->next())
			i->inheritTagsOf(before);

	preparePlug(item);

	Item *last = item->lastSibling();
	if (before) {
		// The normal case:
		for (Item *i = item; i; i = i->next())
			i->setParentItem(before->parentItem());
		item->setPrev(before->prev());
		last->setNext(before);
		before->setPrev(last);
		if (item->prev())
			item->prev()->setNext(item);
		else {
			if (item->parentItem())
				item->parentItem()->setFirstChild(item);
			else
				m_firstItem = item;
		}
	} else {
		// There is no item in the basket:
		for (Item *i = item; i; i = i->next())
			i->setParentItem(0);
		m_firstItem = item;
//		item->setPrev(0);
//		last->setNext(0);
	}

	finishPlug();
}

DecoratedBasket* Basket::decoration()
{
	return (DecoratedBasket*)parent();
}

void Basket::preparePlug(Item *item)
{
	// Select only the new items, compute the new items count and the new number of found items:
	unselectAll();
	int count  = 0;
	int founds = 0;
	for (Item *i = item; i; i = i->next()) {
		if (m_loaded)
			i->setSelectedRecursivly(true); // Items should have a parent basket (and they have, so that's OK).
		count  += i->count();
		founds += i->newSearch(decoration()->searchData());
	}
	m_count += count;
	m_countFounds += founds;

	// If some items don't match (are hidden), tell it to the user:
	if (m_loaded && founds < count) {
		if      (count == 1)          postMessage( i18n("The new item do not match the filter and is not shown.")    );
		else if (founds == count - 1) postMessage( i18n("A new item do not match the filter and is not shown.")      );
		else if (founds > 0)          postMessage( i18n("Some new items do not match the filter and are not shown.") );
		else                          postMessage( i18n("The new items do not match the filter and are not shown.")  );
	}
}

void Basket::finishPlug()
{
	if (m_loaded)
		signalCountsChanged();
}

void Basket::unplugItem(Item *item, bool willBeReplugged)
{
	// If there is nothing to do...
	if (!item)
		return;

	if (!willBeReplugged) {
		item->setSelectedRecursivly(false); // To removeSelectedItem() and decrease the selectedsCount.
		m_count -= item->count();
		m_countFounds -= item->newSearch(decoration()->searchData());
		signalCountsChanged();
	}

	// If it was the first item, change the first item:
	if (m_firstItem == item)
		m_firstItem = item->next();

	// Change previous and next items:
	if (item->prev())
		item->prev()->setNext(item->next());
	if (item->next())
		item->next()->setPrev(item->prev());

	if (item->parentItem()) {
		// If it was the first item of a group, change the first item of the group:
		if (item->parentItem()->firstChild() == item)
			item->parentItem()->setFirstChild( item->next() );

		if (!item->parentItem()->isColumn()) {
			// Ungroup if still 0 item inside parent group:
			if ( ! item->parentItem()->firstChild() )
				unplugItem(item->parentItem()); // TODO delete

			// Ungroup if still 1 item inside parent group:
			else if ( ! item->parentItem()->firstChild()->next() )
				ungroupItem(item->parentItem());
		}
	}

	item->setParentItem(0);
	item->setPrev(0);
	item->setNext(0);

//	recomputeBlankRects(); // FIXME: called too much time. It's here because when dragging and moving an item to another basket and then go back to the original basket, the item is deleted but the item rect is not painter anymore.
}

void Basket::ungroupItem(Item *group)
{
	Item *item            = group->firstChild();
	Item *lastGroupedItem = group;
	Item *nextItem;

	// Move all items after the group (not before, to avoid to change m_firstItem or group->m_firstChild):
	while (item) {
		nextItem = item->next();
//		unplugItem(item, /*willBeReplugged=*/true);
		item->setParentItem(0);
		item->setPrev(0);
		item->setNext(0);
		appendItemAfter(item, lastGroupedItem, /*alreadyInBasket=*/true);
		item->setGroupWidth(group->groupWidth() - Item::GROUP_WIDTH);
		lastGroupedItem = item;
		item = nextItem;
	}
	group->setFirstChild(0);

	// Unplug the group:
	unplugItem(group); // TODO: delete

	relayoutItems(true);
}

void Basket::groupItemBefore(Item *item, Item *with)
{
	if (!item || !with)
		// No item to group or nowhere to group it:
		return;

//	if (m_loaded && !item->isPrimary())
		for (Item *i = item; i; i = i->next())
			i->inheritTagsOf(with);

	preparePlug(item);

	Item *last = item->lastSibling();

	Item *group = new Item(this);
	group->setPrev(with->prev());
	group->setNext(with->next());
	group->setX(with->x());
	group->setY(with->y());
	if (with->parentItem() && with->parentItem()->firstChild() == with)
		with->parentItem()->setFirstChild(group);
	else if (m_firstItem == with)
		m_firstItem = group;
	group->setParentItem(with->parentItem());
	group->setFirstChild(item);
	group->setGroupWidth(with->groupWidth() + Item::GROUP_WIDTH);

	if (with->prev())
		with->prev()->setNext(group);
	if (with->next())
		with->next()->setPrev(group);
	with->setParentItem(group);
	with->setPrev(last);
	with->setNext(0L);

	for (Item *i = item; i; i = i->next())
		i->setParentItem(group);
//	item->setPrev(0L);
	last->setNext(with);

	finishPlug();
}

void Basket::groupItemAfter(Item *item, Item *with)
{
	if (!item || !with)
		// No item to group or nowhere to group it:
		return;

//	if (m_loaded && !item->isPrimary())
		for (Item *i = item; i; i = i->next())
			i->inheritTagsOf(with);

	preparePlug(item);

//	Item *last = item->lastSibling();

	Item *group = new Item(this);
	group->setPrev(with->prev());
	group->setNext(with->next());
	group->setX(with->x());
	group->setY(with->y());
	if (with->parentItem() && with->parentItem()->firstChild() == with)
		with->parentItem()->setFirstChild(group);
	else if (m_firstItem == with)
		m_firstItem = group;
	group->setParentItem(with->parentItem());
	group->setFirstChild(with);
	group->setGroupWidth(with->groupWidth() + Item::GROUP_WIDTH);

	if (with->prev())
		with->prev()->setNext(group);
	if (with->next())
		with->next()->setPrev(group);
	with->setParentItem(group);
	with->setPrev(0L);
	with->setNext(item);

	for (Item *i = item; i; i = i->next())
		i->setParentItem(group);
	item->setPrev(with);
//	last->setNext(0L);

	finishPlug();

	/*if (m_loaded && !item->isFree()/ * && !item->isColumns()* /)
		item->inheritTagsOf(with);*/
}

void Basket::loadItems(const QDomElement &items, Item *parent)
{
	Item *item;
	for (QDomNode n = items.firstChild(); !n.isNull(); n = n.nextSibling()) {
		QDomElement e = n.toElement();
		if (e.isNull()) // Cannot handle that!
			continue;
		item = 0;
		// Load a Group:
		if (e.tagName() == "group") {
			item = new Item(this);      // 1. Create the group...
			loadItems(e, item);         // 3. ... And populate it with child items.
			int itemCount = item->count();
			if (itemCount > 0 || (parent == 0 && !isFreeLayout())) { // But don't remove columns!
				appendItemIn(item, parent); // 2. ... Insert it... FIXME: Initially, the if() the insrtion was the step 2. Was it on purpose?
				// The items in the group are counted two times (it's why appendItemIn() was called before loadItems):
				m_count       -= itemCount;// TODO: Recompute item count every time itemCount() is emitted!
				m_countFounds -= itemCount;
			}
		}
		// Load a Content-Based Item:
		if (e.tagName() == "item") {
			item = new Item(this);      // Create the item...
			ItemFactory__loadNode(XMLWork::getElement(e, "content"), e.attribute("type"), item); // ... Populate it with content...
			appendItemIn(item, parent); // ... And insert it.
			// Load dates:
			if (e.hasAttribute("added"))
				item->setAddedDate(           QDateTime::fromString(e.attribute("added"),            Qt::ISODate));
			if (e.hasAttribute("lastModification"))
				item->setLastModificationDate(QDateTime::fromString(e.attribute("lastModification"), Qt::ISODate));
		}
		// If we successfully loaded an item:
		if (item) {
			// Free Item Properties:
			if (item->isFree()) {
				int x = e.attribute("x").toInt();
				int y = e.attribute("y").toInt();
				item->setX(x < 0 ? 0 : x);
				item->setY(y < 0 ? 0 : y);
			}
			// Resizeable Item Properties:
			if (item->hasResizer() || item->isColumn())
				item->setGroupWidth(e.attribute("width", "200").toInt());
			// Group Properties:
			if (item->isGroup() && !item->isColumn() && XMLWork::trueOrFalse(e.attribute("folded", "false")))
				item->toggleFolded(false);
			// Tags:
			if (item->content()) {
				QString tagsString = XMLWork::getElementText(e, "tags", "");
				QStringList tagsId = QStringList::split(";", tagsString);
				for (QStringList::iterator it = tagsId.begin(); it != tagsId.end(); ++it) {
					State *state = Tag::stateForId(*it);
					if (state)
						item->addState(state, /*orReplace=*/true);
				}
			}
		}
	}
}

void Basket::saveItems(QDomDocument &document, QDomElement &element, Item *parent)
{
	Item *item = (parent ? parent->firstChild() : firstItem());
	while (item) {
		// Create Element:
		QDomElement itemElement = document.createElement(item->isGroup() ? "group" : "item");
		element.appendChild(itemElement);
		// Free Item Properties:
		if (item->isFree()) {
			itemElement.setAttribute("x", item->finalX());
			itemElement.setAttribute("y", item->finalY());
		}
		// Resizeable Item Properties:
		if (item->hasResizer())
			itemElement.setAttribute("width", item->groupWidth());
		// Group Properties:
		if (item->isGroup() && !item->isColumn())
			itemElement.setAttribute("folded", XMLWork::trueOrFalse(item->isFolded()));
		// Save Content:
		if (item->content()) {
			// Save Dates:
			itemElement.setAttribute("added",            item->addedDate().toString(Qt::ISODate)           );
			itemElement.setAttribute("lastModification", item->lastModificationDate().toString(Qt::ISODate));
			// Save Content:
			itemElement.setAttribute("type", item->content()->lowerTypeName());
			QDomElement content = document.createElement("content");
			itemElement.appendChild(content);
			item->content()->saveToNode(document, content);
			// Save Tags:
			if (item->states().count() > 0) {
				QString tags;
				for (State::List::iterator it = item->states().begin(); it != item->states().end(); ++it)
					tags += (tags.isEmpty() ? "" : ";") + (*it)->id();
				XMLWork::addElement(document, itemElement, "tags", tags);
			}
		} else
			// Save Child Items:
			saveItems(document, itemElement, item);
		// Go to the Next One:
		item = item->next();
	}
}

void Basket::loadProperties(const QDomElement &properties)
{
//	setLocked(          XMLWork::trueOrFalse(XMLWork::getElementText(properties, "locked",         "false"))     );
//	setShowSearchBar(   XMLWork::trueOrFalse(XMLWork::getElementText(properties, "showSearchBar",  "false"))     );

	// Compute Default Values for When Loading the Properties:
	QString defaultBackgroundColor = (backgroundColorSetting().isValid() ? backgroundColorSetting().name() : "");
	QString defaultTextColor       = (textColorSetting().isValid()       ? textColorSetting().name()       : "");

	// Load the Properties:
	QString icon = XMLWork::getElementText(properties, "icon", this->icon() );
	QString name = XMLWork::getElementText(properties, "name", basketName() );

	QDomElement appearance = XMLWork::getElement(properties, "appearance");
	QString backgroundImage       = appearance.attribute( "backroundImage", backgroundImageName()  );
	QString backgroundColorString = appearance.attribute( "backroundColor", defaultBackgroundColor );
	QString textColorString       = appearance.attribute( "textColor",      defaultTextColor       );
	QColor  backgroundColor = (backgroundColorString.isEmpty() ? QColor() : QColor(backgroundColorString));
	QColor  textColor       = (textColorString.isEmpty()       ? QColor() : QColor(textColorString)      );

	QDomElement disposition = XMLWork::getElement(properties, "disposition");
	bool free        = XMLWork::trueOrFalse( disposition.attribute( "free",        XMLWork::trueOrFalse(isFreeLayout())   ) );
	int  columnCount =                       disposition.attribute( "columnCount", QString::number(this->columnsCount())  ).toInt();
	bool mindMap     = XMLWork::trueOrFalse( disposition.attribute( "mindMap",     XMLWork::trueOrFalse(isMindMap())      ) );

	QDomElement shortcut = XMLWork::getElement(properties, "shortcut");
	QString actionStrings[] = { "show", "globalShow", "globalSwitch" };
	KShortcut combination  = KShortcut( shortcut.attribute( "combination", m_action->shortcut().toStringInternal() ) );
	QString   actionString =            shortcut.attribute( "action" );
	int action = shortcutAction();
	if (actionString == actionStrings[0]) action = 0;
	if (actionString == actionStrings[1]) action = 1;
	if (actionString == actionStrings[2]) action = 2;

	// Apply the Properties:
	setDisposition((free ? (mindMap ? 2 : 1) : 0), columnCount);
	setShortcut(combination, action);
	setAppearance(icon, name, backgroundImage, backgroundColor, textColor); // Will emit propertiesChanged(this)
}

void Basket::saveProperties(QDomDocument &document, QDomElement &properties)
{
	XMLWork::addElement( document, properties, "name", basketName() );
	XMLWork::addElement( document, properties, "icon", icon()       );

	QDomElement appearance = document.createElement("appearance");
	properties.appendChild(appearance);
	appearance.setAttribute( "backroundImage", backgroundImageName() );
	appearance.setAttribute( "backroundColor", backgroundColorSetting().isValid() ? backgroundColorSetting().name() : "" );
	appearance.setAttribute( "textColor",      textColorSetting().isValid()       ? textColorSetting().name()       : "" );

	QDomElement disposition = document.createElement("disposition");
	properties.appendChild(disposition);
	disposition.setAttribute( "free",        XMLWork::trueOrFalse(isFreeLayout()) );
	disposition.setAttribute( "columnCount", QString::number(columnsCount())      );
	disposition.setAttribute( "mindMap",     XMLWork::trueOrFalse(isMindMap())    );

	QDomElement shortcut = document.createElement("shortcut");
	properties.appendChild(shortcut);
	QString actionStrings[] = { "show", "globalShow", "globalSwitch" };
	shortcut.setAttribute( "combination", m_action->shortcut().toStringInternal() );
	shortcut.setAttribute( "action",      actionStrings[shortcutAction()]         );
}

void Basket::subscribeBackgroundImages()
{
	if (!m_backgroundImageName.isEmpty()) {
		Global::backgroundsManager->subscribe(m_backgroundImageName);
		Global::backgroundsManager->subscribe(m_backgroundImageName, this->backgroundColor());
		Global::backgroundsManager->subscribe(m_backgroundImageName, selectionRectInsideColor());
		m_backgroundPixmap         = Global::backgroundsManager->pixmap(m_backgroundImageName);
		m_opaqueBackgroundPixmap   = Global::backgroundsManager->opaquePixmap(m_backgroundImageName, this->backgroundColor());
		m_selectedBackgroundPixmap = Global::backgroundsManager->opaquePixmap(m_backgroundImageName, selectionRectInsideColor());
		m_backgroundTiled          = Global::backgroundsManager->tiled(m_backgroundImageName);
	}
}

void Basket::unsubscribeBackgroundImages()
{
	if (hasBackgroundImage()) {
		Global::backgroundsManager->unsubscribe(m_backgroundImageName);
		Global::backgroundsManager->unsubscribe(m_backgroundImageName, this->backgroundColor());
		Global::backgroundsManager->unsubscribe(m_backgroundImageName, selectionRectInsideColor());
		m_backgroundPixmap         = 0;
		m_opaqueBackgroundPixmap   = 0;
		m_selectedBackgroundPixmap = 0;
	}
}

void Basket::setAppearance(const QString &icon, const QString &name, const QString &backgroundImage, const QColor &backgroundColor, const QColor &textColor)
{
	unsubscribeBackgroundImages();

	m_icon                   = icon;
	m_basketName             = name;
	m_backgroundImageName    = backgroundImage;
	m_backgroundColorSetting = backgroundColor;
	m_textColorSetting       = textColor;

	// Basket should ALWAYS have an icon (the "basket" icon by default):
	QPixmap iconTest = kapp->iconLoader()->loadIcon(m_icon, KIcon::NoGroup, 16, KIcon::DefaultState, 0L, /*canReturnNull=*/true);
	if (iconTest.isNull())
		m_icon = "basket";

	// We don't request the background images if it's not loaded yet (to make the application startup fast).
	// When the basket is loading (because requested by the user: he/she want to access it)
	// it load the properties, subscribe to (and then load) the images, update the "Loading..." message with the image,
	// load all the items and it's done!
//	if (m_loaded)
//		subscribeBackgroundImages();
	if (m_loadingLaunched)
		subscribeBackgroundImages();

	recomputeAllStyles(); // If an item have a tag with the same background color as the basket one, then display a "..."
	recomputeBlankRects(); // See the drawing of blank areas in Basket::drawContents()
	unbufferizeAll();
	updateContents();

	emit propertiesChanged(this);
}

void Basket::setDisposition(int disposition, int columnCount)
{
	static const int COLUMNS_LAYOUT  = 0;
	static const int FREE_LAYOUT     = 1;
	static const int MINDMAPS_LAYOUT = 2;

	int currentDisposition = (isFreeLayout() ? (isMindMap() ? MINDMAPS_LAYOUT : FREE_LAYOUT) : COLUMNS_LAYOUT);

	if (currentDisposition == COLUMNS_LAYOUT && disposition == COLUMNS_LAYOUT) {
		if (firstItem() && columnCount > m_columnsCount) {
			// Insert each new columns:
			for (int i = m_columnsCount; i < columnCount; ++i) {
				Item *newColumn = new Item(this);
				insertItem(newColumn, /*clicked=*/lastItem(), /*zone=*/Item::BottomInsert, QPoint(), /*animateNewPosition=*/false);
			}
		} else if (firstItem() && columnCount < m_columnsCount) {
			Item *column = firstItem();
			Item *cuttedItems = 0;
			for (int i = 1; i <= m_columnsCount; ++i) {
				Item *columnToRemove = column;
				column = column->next();
				if (i > columnCount) {
					// Remove the columns that are too much:
					unplugItem(columnToRemove, /*willBeReplugged=*/false);
					// "Cut" the content in the columns to be deleted:
					if (columnToRemove->firstChild()) {
						for (Item *it = columnToRemove->firstChild(); it; it = it->next())
							it->setParentItem(0);
						if (!cuttedItems)
							cuttedItems = columnToRemove->firstChild();
						else {
							Item *lastCuttedItem = cuttedItems;
							while (lastCuttedItem->next())
								lastCuttedItem = lastCuttedItem->next();
							lastCuttedItem->setNext(columnToRemove->firstChild());
							columnToRemove->firstChild()->setPrev(lastCuttedItem);
						}
						columnToRemove->setFirstChild(0);
					}
				}
			}
			// Paste the content in the last column:
			if (cuttedItems)
				insertItem(cuttedItems, /*clicked=*/lastItem(), /*zone=*/Item::BottomColumn, QPoint(), /*animateNewPosition=*/true);
			unselectAll();
		}
		if (columnCount != m_columnsCount) {
			m_columnsCount = (columnCount <= 0 ? 1 : columnCount);
			equalizeColumnSizes(); // Will relayoutItems()
		}
	} else if (currentDisposition == COLUMNS_LAYOUT && (disposition == FREE_LAYOUT || disposition == MINDMAPS_LAYOUT)) {
		Item *column = firstItem();
		m_columnsCount = 0; // Now, so relayoutItems() will not relayout the free items as if they were columns!
		while (column) {
			// Move all childs on the first level:
			Item *nextColumn = column->next();
			ungroupItem(column);
			column = nextColumn;
		}
		unselectAll();
		m_mindMap = (disposition == MINDMAPS_LAYOUT);
		relayoutItems(true);
	} else if ((currentDisposition == FREE_LAYOUT || currentDisposition == MINDMAPS_LAYOUT) && disposition == COLUMNS_LAYOUT) {
		if (firstItem()) {
			// TODO: Reorder items!
			// Remove all items (but keep a reference to them, we're not crazy ;-) ):
			Item *items = m_firstItem;
			m_firstItem = 0;
			m_count = 0;
			m_countFounds = 0;
			// Insert the number of columns that is needed:
			Item *lastInsertedColumn = 0;
			for (int i = 0; i < columnCount; ++i) {
				Item *column = new Item(this);
				if (lastInsertedColumn)
					insertItem(column, /*clicked=*/lastInsertedColumn, /*zone=*/Item::BottomInsert, QPoint(), /*animateNewPosition=*/false);
				else
					m_firstItem = column;
				lastInsertedColumn = column;
			}
			// Reinsert the old items in the first column:
			insertItem(items, /*clicked=*/firstItem(), /*zone=*/Item::BottomColumn, QPoint(), /*animateNewPosition=*/true);
			unselectAll();
		}
		m_columnsCount = (columnCount <= 0 ? 1 : columnCount);
		equalizeColumnSizes(); // Will relayoutItems()
	}
}

void Basket::equalizeColumnSizes()
{
	if (!firstItem())
		return;

	// Necessary to know the available space;
	relayoutItems(true);

	int availableSpace = visibleWidth();
	int columnWidth = (visibleWidth() - (columnsCount()-1)*Item::GROUP_WIDTH) / columnsCount();
	int columnCount = columnsCount();
	Item *column = firstItem();
	while (column) {
		int minGroupWidth = column->minRight() - column->x();
		if (minGroupWidth > columnWidth) {
			availableSpace -= minGroupWidth;
			--columnCount;
		}
		column = column->next();
	}
	columnWidth = (availableSpace - (columnsCount()-1)*Item::GROUP_WIDTH) / columnCount;

	column = firstItem();
	while (column) {
		int minGroupWidth = column->minRight() - column->x();
		if (minGroupWidth > columnWidth)
			column->setGroupWidth(minGroupWidth);
		else
			column->setGroupWidth(columnWidth);
		column = column->next();
	}

	relayoutItems(true);
}

void Basket::save()
{
	if (!m_loaded)
		return;

	DEBUG_WIN << "Basket[" + folderName() + "]: Saving...";

	// Create Document:
	QDomDocument document(/*doctype=*/"basket");
	QDomElement root = document.createElement("basket");
	document.appendChild(root);

	// Create Properties Element and Populate It:
	QDomElement properties = document.createElement("properties");
	saveProperties(document, properties);
	root.appendChild(properties);

	// Create Items Element and Populate It:
	QDomElement items = document.createElement("items");
	saveItems(document, items, 0);
	root.appendChild(items);

	// Write to Disk:
	QFile file(fullPath() + "/.basket");
	if (file.open(IO_WriteOnly)) {
		QTextStream stream(&file);
		stream.setEncoding(QTextStream::UnicodeUTF8);
		QString xml = document.toString();
		stream << "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n";
		stream << xml;
		file.close();
	} else
		DEBUG_WIN << "Basket[" + folderName() + "]: <font color=red>FAILED</font>!";
}

void Basket::load()
{
	// Load only once:
	if (m_loadingLaunched)
		return;
	m_loadingLaunched = true;

//	m_loadingFramesTimer.start(50/*ms*/);

	DEBUG_WIN << "Basket[" + folderName() + "]: Loading...";

	QDomDocument *doc = XMLWork::openFile("basket", fullPath() + "/.basket");
	if (doc == 0) {
		DEBUG_WIN << "Basket[" + folderName() + "]: <font color=red>FAILED</font>!";
		return;
	}

	QDomElement docElem = doc->documentElement();
	QDomElement properties = XMLWork::getElement(docElem, "properties");

	// Hack: loadProperties() will call setAppearance() with the new background image.
	// Background images are effectively loaded from image files only when m_loaded is true.
	// So we set that variable temporarily:
//	m_loaded = true;
	loadProperties(properties);
//	m_loaded = false;
	// Now that the background image is loaded and subscribed, we display it during the load process:
	updateContents();
	kapp->processEvents();

	loadItems(XMLWork::getElement(docElem, "items"), 0L);
	signalCountsChanged();
	if (isColumnsLayout()) {
		// Count the number of columns:
		int columnsCount = 0;
		Item *column = firstItem();
		while (column) {
			++columnsCount;
			column = column->next();
		}
		m_columnsCount = columnsCount;
	}

	relayoutItems(false);
	if (Settings::playAnimations() && !decoration()->searchBar()->searchData().isSearching) // No animation when filtering all!
		animateLoad();//QTimer::singleShot( 0, this, SLOT(animateLoad()) );
	else
		m_loaded = true;
//	m_loadingFramesTimer.stop();
}

/*Item* Basket::nextShownItemFrom(Item *item, int step)
{
	if (step > 0) {
		while (step && item) {
			item = item->next();
			if (item && item->isShown())
				step--;
		}
	} else {
		while (step && item) {
			item = item->previous();
			if (item && item->isShown())
				step++;
		}
	}
	return item;
}*/

void Basket::searchAgain()
{
	newSearch( ((DecoratedBasket*)parent())->searchData() );
}

void Basket::newSearch(const SearchData &data)
{
	m_countFounds = 0;
	for (Item *item = firstItem(); item; item = item->next())
		m_countFounds += item->newSearch(data);

	relayoutItems(true);
	doHoverEffects();
	signalCountsChanged();

//	if (hasFocus())    // if (!hasFocus()), focusAnItem() will be called at focusInEvent()
//		focusAnItem(); //  so, we avoid de-focus an item if it will be re-shown soon
//	if (m_focusedItem != 0L)
//		ensureVisibleItem(m_focusedItem);

//	Global::mainContainer->setSearching(true);
}

QString Basket::fullPath()
{
	return Global::basketsFolder() + folderName();
}

QString Basket::fullPathForFileName(const QString &fileName)
{
	return fullPath() + fileName;
}

/*static*/ QString Basket::fullPathForFolderName(const QString &folderName)
{
	return Global::basketsFolder() + folderName;
}


void Basket::setShortcut(KShortcut shortcut, int action)
{
	QString sAction = "global_basket_activate_" + folderName();
//	if (!Global::globalAccel->shortcut(sAction).isNull()) {
		Global::globalAccel->remove(sAction);
		Global::globalAccel->updateConnections();
//	}

	m_action->setShortcut(shortcut);
	m_shortcutAction = action;

	if (action > 0)// {
		Global::globalAccel->insert(sAction, m_action->text(), /*WhatsThis=*/"", m_action->shortcut(), KShortcut(), this, SLOT(activatedShortcut()), /*configurable=*/false);
	Global::globalAccel->updateConnections();
	//}
}

void Basket::activatedShortcut()
{
	Global::mainContainer->setCurrentBasket(this);

	if (m_shortcutAction == 1)
		Global::mainContainer->setActive(true);
}



void Basket::signalCountsChanged()
{
	if (!m_timerCountsChanged.isActive())
		m_timerCountsChanged.start(0/*ms*/, /*singleShot=*/true);
}

void Basket::countsChangedTimeOut()
{
	emit countsChanged(this);
}


Basket::Basket(QWidget *parent, const QString &folderName)
 : QScrollView(parent),
   QToolTip(viewport()),
   m_noActionOnMouseRelease(false), m_pressPos(-100, -100), m_canDrag(false),
   m_firstItem(0), m_columnsCount(1), m_mindMap(false), m_resizingItem(0L), m_pickedResizer(0), m_movingItem(0L), m_pickedHandle(0, 0),
   m_clickedToInsert(0), m_zoneToInsert(0), m_posToInsert(-1, -1),
   m_loaded(false), m_loadingLaunched(false),
   m_backgroundPixmap(0), m_opaqueBackgroundPixmap(0), m_selectedBackgroundPixmap(0),
   m_action(0), m_shortcutAction(0),
   m_hoveredItem(0), m_lockedHovering(false), m_underMouse(false),
   m_inserterRect(), m_inserterShown(false), m_inserterSplit(true), m_inserterTop(false), m_inserterGroup(false),
   m_isSelecting(false), m_selectionStarted(false),
   m_count(0), m_countFounds(0), m_countSelecteds(0),
   m_folderName(folderName),
   m_editor(0), m_redirectEditActions(false),
   m_isDuringDrag(false), m_draggedItems()
{
	QString sAction = "local_basket_activate_" + folderName;
	m_action = new KAction("FAKE TEXT", "FAKE ICON", KShortcut(), this, SLOT(activatedShortcut()), Global::mainContainer->actionCollection(), sAction);
	m_action->setShortcutConfigurable(false); // We do it in the basket properties dialog (and keep it in sync with the global one)

	if (!m_folderName.endsWith("/"))
		m_folderName += "/";

	setFocusPolicy(QWidget::StrongFocus);
//	setFrameShape(QFrame::NoFrame);
	setWFlags(Qt::WNoAutoErase);
	setDragAutoScroll(true);

	viewport()->setAcceptDrops(true);
	viewport()->setMouseTracking(true);
	viewport()->setBackgroundMode(NoBackground); // Do not clear the widget before paintEvent() because we always draw every pixels (faster and flicker-free)

	connect( &m_animationTimer,           SIGNAL(timeout()),   this, SLOT(animateObjects())        );
	connect( &m_autoScrollSelectionTimer, SIGNAL(timeout()),   this, SLOT(doAutoScrollSelection()) );
	connect( &m_timerCountsChanged,       SIGNAL(timeout()),   this, SLOT(countsChangedTimeOut())  );
//	connect(  m_action,                   SIGNAL(activated()), this, SLOT(activatedShortcut())     );
//	connect( &m_loadingFramesTimer,       SIGNAL(timeout()),   this, SLOT(updateLoadingFrame())    );
	connect( this, SIGNAL(contentsMoving(int, int)), this, SLOT(contentsMoved()) );
}

void Basket::contentsMoved()
{
	// This slot is called BEFORE the content move, so we delay the hover effects:
	QTimer::singleShot(0, this, SLOT(doHoverEffects()));
}

/*void Basket::updateLoadingFrame()
{
	++m_loadingFrames;

	static const int showDelay = 0/ *ms* /;
	static const int fadeDelay = 500/ *ms* /;
	static const int frameRate = 50/ *ms* /;

	std::cout << "updateLoadingFrame()" << std::endl;
	// If we are in the fading stage, update the view:
	if (m_loadingFrames >= showDelay / frameRate) {
		std::cout << "updateLoadingFrame():Do It" << std::endl;
		QSimpleRichText rt(QString("<center>%1</center>").arg(i18n("Loading...")), QScrollView::font());
		rt.setWidth(visibleWidth());
		int hrt = rt.height();
		updateContents(0, (visibleHeight() - hrt) / 2, visibleWidth(), hrt);
		// If the fading is finished, stop the effect!:
		if (m_loadingFrames >= (showDelay + fadeDelay) / frameRate) {
			std::cout << "updateLoadingFrame() : Stop It" << std::endl;
			m_loadingFramesTimer.stop();
		}
	}
}*/

void Basket::enterEvent(QEvent *)
{
	m_underMouse = true;
	doHoverEffects();
}

void Basket::leaveEvent(QEvent *)
{
	m_underMouse = false;
	doHoverEffects();

	if (m_lockedHovering)
		return;

	removeInserter();
	if (m_hoveredItem) {
		m_hoveredItem->setHovered(false);
		m_hoveredItem->setHoveredZone(Item::None);
		updateItem(m_hoveredItem);
	}
	m_hoveredItem = 0;
}

void Basket::contentsMousePressEvent(QMouseEvent *event)
{
	QTimer::singleShot(0, this, SLOT(setFocus()));

	// Is during edition, clicking the basket end the editing and do NOTHING ELSE:
	if (isDuringEdit()) {
		closeEditor();
		//m_noActionOnMouseRelease = true;
		//doHoverEffects();
		return;
	}

	// Do nothing if we disabled the click some milliseconds before.
	// For instance when a popup menu has been closed with click, we should not do action:
	if (event->button() == Qt::LeftButton && (kapp->activePopupWidget() || m_lastDisableClick.msecsTo(QTime::currentTime()) <= 80)) {
		doHoverEffects();
		m_noActionOnMouseRelease = true;
		// But we allow to select:
		// The code is the same as at the bottom of this method:
		if (event->button() == Qt::LeftButton) {
			m_selectionStarted = true;
			m_selectionBeginPoint = event->pos();
			m_selectionInvert = event->stateAfter() & Qt::ControlButton || event->stateAfter() & Qt::ShiftButton;
		}
		return;
	}

	Item *clicked = itemAt(event->pos().x(), event->pos().y());
	Item::Zone zone = (clicked ? clicked->zoneAt( event->pos() - QPoint(clicked->x(), clicked->y()) ) : Item::None);

	if (zone == Item::TagsArrow) {
		if (!clicked->isSelected())
			unselectAllBut(clicked);
		m_noActionOnMouseRelease = true;
		popupTagsMenu(clicked);
		return;
	}

	if (event->button() == Qt::LeftButton) {
		// Drag and drop:
		if ( (zone == Item::Handle || zone == Item::Group) ||
		     (clicked && clicked->isSelected() &&
		       (zone == Item::TagsArrow || zone == Item::Custom0 || zone == Item::Content || zone == Item::Link /**/ || zone >= Item::Emblem0 /**/)) ) {
			m_pressPos = event->/*globalPos*/pos(); // TODO: Allow to drag emblems to assign them to other items. Then don't allow drag at Emblem0!!
			m_canDrag  = true;
		}

		// Initializing Resizer move:
		if (zone == Item::Resizer) {
			m_resizingItem  = clicked;
			m_pickedResizer = event->pos().x() - clicked->rightLimit();
			m_noActionOnMouseRelease = true;
			m_lockedHovering = true;
			return;
		}

		// Select item(s):
		if (zone == Item::Handle || zone == Item::Group) {
			if (event->stateAfter() & Qt::ControlButton)
				clicked->setSelectedRecursivly(!clicked->allSelected());
			else if (!clicked->allSelected())
				unselectAllBut(clicked);
		} else if (clicked && (event->stateAfter() & Qt::ControlButton) && zone != Item::None && zone != Item::BottomColumn && zone != Item::Resizer) {
			clicked->setSelectedRecursivly(!clicked->allSelected());
			m_noActionOnMouseRelease = true;
			return;
		}

		// Initializing Item move:
/*		if ((zone == Item::Group || zone == Item::Handle) && clicked->isFree()) {
			m_movingItem   = clicked;
			m_pickedHandle = QPoint(event->pos().x() - clicked->x(), event->pos().y() - clicked->y());
			m_noActionOnMouseRelease = true;
			m_lockedHovering = true;
			return;
		}
*/

		// Folding/Unfolding group:
		if (zone == Item::GroupExpander) {
			clicked->toggleFolded(Settings::playAnimations());
			relayoutItems(true);
			m_noActionOnMouseRelease = true;
			return;
		}
	}

	// Popup menu for tag emblems:
	if (event->button() == Qt::RightButton && zone >= Item::Emblem0) {
		if (!clicked->isSelected())
			unselectAllBut(clicked);
		popupEmblemMenu(clicked, zone - Item::Emblem0);
		m_noActionOnMouseRelease = true;
		return;
	}

	// Insertion Popup Menu:
	if ( (event->button() == Qt::RightButton) &&
	     ((!clicked && isFreeLayout()) ||
	      (clicked && (zone == Item::TopInsert || zone == Item::TopGroup || zone == Item::BottomInsert || zone == Item::BottomGroup || zone == Item::BottomColumn))) ) {
		unselectAll();
		m_clickedToInsert = clicked;
		m_zoneToInsert    = zone;
		m_posToInsert     = event->pos();
		KPopupMenu* menu = (KPopupMenu*)(Global::mainContainer->popupMenu("insert_popup"));
		if (!menu->title(/*id=*/120).isEmpty()) // If we already added a title, remove it because it would be kept and then added several times:
			menu->removeItem(/*id=*/120);
		menu->insertTitle((zone == Item::TopGroup || zone == Item::BottomGroup ? i18n("The verb (Group New Item)", "Group") : i18n("The verb (Insert New Item)", "Insert")), /*id=*/120, /*index=*/0);
		connect( menu, SIGNAL(aboutToHide()),  this, SLOT(unlockHovering())      );
		connect( menu, SIGNAL(aboutToHide()),  this, SLOT(disableNextClick())    );
		connect( menu, SIGNAL(aboutToHide()),  this, SLOT(hideInsertPopupMenu()) );
		doHoverEffects(clicked, zone); // In the case where another popup menu was open, we should do that manually!
		m_lockedHovering = true;
		menu->exec(QCursor::pos());
		m_noActionOnMouseRelease = true;
		return;
	}

	// Item Context Menu:
	if (event->button() == Qt::RightButton && clicked && !clicked->isColumn() && zone != Item::Resizer) {
		if (!clicked->isSelected())
			unselectAllBut(clicked);
		QPopupMenu* menu = Global::mainContainer->popupMenu("item_popup");
		connect( menu, SIGNAL(aboutToHide()),  this, SLOT(unlockHovering())   );
		connect( menu, SIGNAL(aboutToHide()),  this, SLOT(disableNextClick()) );
		doHoverEffects(clicked, zone); // In the case where another popup menu was open, we should do that manually!
		m_lockedHovering = true;
		menu->exec(QCursor::pos());
		m_noActionOnMouseRelease = true;
		return;
	}

	// Paste selection under cursor (but not "create new primary item under cursor" because this is on moveRelease):
	if (event->button() == Qt::MidButton && zone != Item::Resizer) {
		if (clicked)
			zone = clicked->zoneAt( event->pos() - QPoint(clicked->x(), clicked->y()), true );
		clickedToInsert(event, clicked, zone);
		m_noActionOnMouseRelease = true;
		save();
		return;
	}

	// Finally, no action has been done durint pressEvent, so an action can be done on releaseEvent:
	m_noActionOnMouseRelease = false;

	/* Selection scenario:
	 * On contentsMousePressEvent, put m_selectionStarted to true and init Begin and End selection point.
	 * On contentsMouseMoveEvent, if m_selectionStarted, update End selection point, update selection rect,
	 * and if it's larger, switching to m_isSelecting mode: we can draw the selection rectangle.
	 */
	// Prepare selection:
	if (event->button() == Qt::LeftButton) {
		m_selectionStarted = true;
		m_selectionBeginPoint = event->pos();
		// We usualy invert the selection with the Ctrl key, but some environements (like GNOME or The Gimp) do it with the Shift key.
		// Since the Shift key has no specific usage, we allow to invert selection ALSO with Shift for Gimp people
		m_selectionInvert = event->stateAfter() & Qt::ControlButton || event->stateAfter() & Qt::ShiftButton;
	}
}

void Basket::disableNextClick()
{
	m_lastDisableClick = QTime::currentTime();
}

void Basket::recomputeAllStyles()
{
	FOR_EACH_ITEM (item)
		item->recomputeAllStyles();
}

void Basket::insertItem(Item *item, Item *clicked, int zone, const QPoint &pos, bool animateNewPosition)
{
	if (!item) {
		std::cout << "Wanted to insert NO item" << std::endl;
		return;
	}

	if (clicked && zone == Item::BottomColumn) {
		// When inserting at the bottom of a column, it's obvious the new item SHOULD inherit tags.
		// We ensure that by changing the insertion point after the last item of the column:
		Item *last = clicked->lastChild();
		if (last) {
			clicked = last;
			zone = Item::BottomInsert;
		}
	}

	/// Insertion at the bottom of a column:
	if (clicked && zone == Item::BottomColumn) {
		item->setWidth(clicked->rightLimit() - clicked->x());
		Item *lastChild = clicked->lastChild();
		if (!animateNewPosition || !Settings::playAnimations())
			for (Item *i = item; i; i = i->next()) {
				i->setXRecursivly(clicked->x());
				i->setYRecursivly((lastChild ? lastChild : clicked)->bottom() + 1);
			}
		appendItemIn(item, clicked);

	/// Insertion relative to an item (top/bottom, insert/group):
	} else if (clicked) {
		item->setWidth(clicked->width());
		if (!animateNewPosition || !Settings::playAnimations())
			for (Item *i = item; i; i = i->next()) {
				if (zone == Item::TopGroup || zone == Item::BottomGroup)
					i->setXRecursivly(clicked->x() + Item::GROUP_WIDTH);
				else
					i->setXRecursivly(clicked->x());
				if (zone == Item::TopInsert || zone == Item::TopGroup)
					i->setYRecursivly(clicked->y());
				else
					i->setYRecursivly(clicked->bottom() + 1);
			}
		if      (zone == Item::TopInsert)    { appendItemBefore(item, clicked); }
		else if (zone == Item::BottomInsert) { appendItemAfter(item,  clicked); }
		else if (zone == Item::TopGroup)     { groupItemBefore(item,  clicked); }
		else if (zone == Item::BottomGroup)  { groupItemAfter(item,   clicked); }

	/// Free insertion:
	} else if (isFreeLayout()) {
		// Group if item have siblings:
		if (item->next()) {
			Item *group = new Item(this);
			for (Item *i = item; i; i = i->next())
				i->setParentItem(group);
			group->setFirstChild(item);
			item = group;
		}
		// Insert at cursor position:
		const int initialWidth = 250;
		item->setWidth(item->isGroup() ? Item::GROUP_WIDTH : initialWidth);
		if (item->isGroup() && item->firstChild())
			item->setInitialHeight(item->firstChild()->height());
		//item->setGroupWidth(initialWidth);
		if (animateNewPosition && Settings::playAnimations())
			item->setFinalPosition(pos.x(), pos.y());
		else {
			item->setXRecursivly(pos.x());
			item->setYRecursivly(pos.y());
		}
		appendItemAfter(item, lastItem());
	}

	relayoutItems(true);
	doHoverEffects();
}

void Basket::clickedToInsert(QMouseEvent *event, Item *clicked, /*Item::Zone*/int zone)
{
	Item *item;
	if (event->button() == Qt::MidButton)
		item = ItemFactory::dropItem(KApplication::clipboard()->data(QClipboard::Selection), this);
	else
		item = ItemFactory::createItemText("", this);

	if (!item)
		return;

	insertItem(item, clicked, zone, event->pos(), /*animateNewPosition=*/false);

//	ensureVisibleItem(lastInsertedItem());
//	unselectAllBut(lastInsertedItem());
//	setFocusedItem(lastInsertedItem());

	if (event->button() != Qt::MidButton) {
		//		unselectAllBut(item); // TODO: Reenable this
		removeInserter(); // Case: user clicked below a column to insert, the item is inserted and doHoverEffects() put a new inserter below. We don't want it.
		itemEdit(item, /*justAdded=*/true);
	}
}







void Basket::contentsDragEnterEvent(QDragEnterEvent *event)
{
	m_isDuringDrag = true;
	Global::mainContainer->updateStatusBarHint();
	if (ItemDrag::basketOf(event) == this)
		m_draggedItems = ItemDrag::itemsOf(event);
}

void Basket::contentsDragMoveEvent(QDragMoveEvent *event)
{
//	m_isDuringDrag = true;

//	if (isLocked())
//		return;

//	FIXME: viewportToContents does NOT work !!!
//	QPoint pos = viewportToContents(event->pos());
//	QPoint pos( event->pos().x() + contentsX(), event->pos().y() + contentsY() );

//	if (insertAtCursorPos())
//		computeInsertPlace(pos);
	doHoverEffects(event->pos());

//	showFrameInsertTo();
	if (isFreeLayout() || itemAt(event->pos().x(), event->pos().y())) // Cursor before rightLimit() or hovering the dragged source items
		acceptDropEvent(event);
	else {
		event->acceptAction(false);
		event->accept(false);
	}

/*	Item *hoveredItem = itemAt(event->pos().x(), event->pos().y());
	if ( (isColumnsLayout() && !hoveredItem) || (draggedItems().contains(hoveredItem)) ) {
		event->acceptAction(false);
		event->accept(false);
	} else
		acceptDropEvent(event);*/

	// A workarround since QScrollView::dragAutoScroll seem to have no effect :
//	ensureVisible(event->pos().x() + contentsX(), event->pos().y() + contentsY(), 30, 30);
//	QScrollView::dragMoveEvent(event);
}

void Basket::contentsDragLeaveEvent(QDragLeaveEvent*)
{
//	resetInsertTo();
	m_isDuringDrag = false;
	m_draggedItems.clear();
	m_noActionOnMouseRelease = true;
	emit resetStatusBarText();
	doHoverEffects();
}

void Basket::dropEvent(QDropEvent *event)
{
	QScrollView::dropEvent(event);
}

void Basket::contentsDropEvent(QDropEvent *event)
{
	m_isDuringDrag = false;
	m_draggedItems.clear();
	emit resetStatusBarText();

//	if (isLocked())
//		return;

	Item *item = ItemFactory::dropItem( event, this, true, event->action(), dynamic_cast<Item*>(event->source()) );

	if (item) {
		Item *clicked = itemAt(event->pos().x(), event->pos().y());
		Item::Zone zone = (clicked ? clicked->zoneAt( event->pos() - QPoint(clicked->x(), clicked->y()), /*toAdd=*/true ) : Item::None);
		bool animateNewPosition = ItemFactory::movingItemsInTheSameBasket(event, this, event->action());
		if (animateNewPosition) {
			FOR_EACH_ITEM (i)
				i->setOnTop(false);
			// FOR_EACH_ITEM_IN_CHUNK(item)
			for (Item *i = item; i; i = i->next())
				i->setOnTop(true);
		}
		insertItem(item, clicked, zone, event->pos(), animateNewPosition);

//		ensureVisibleItem(lastInsertedItem());
//		unselectAllBut(lastInsertedItem());
//		setFocusedItem(lastInsertedItem());

//		if (event->button() != Qt::MidButton) {
//			unselectAllBut(item);
//			removeInserter(); // Case: user clicked below a column to insert, the item is inserted and doHoverEffects() put a new inserter below. We don't want it.
//		}

//		resetInsertTo();
		doHoverEffects();
		save();
	}
}

void Basket::insertEmptyItem(int type)
{
	if (isDuringEdit())
		closeEditor();
	Item *item = ItemFactory::createEmptyItem((ItemType::Id)type, this);
	insertCreatedItem(item/*, / *edit=* /true*/);
	itemEdit(item, /*justAdded=*/true);

}

void Basket::insertWizard(int type)
{
	saveInsertionData();
	Item *item = 0L;
	switch (type) {
		default:
		case 1: item = ItemFactory::importKMenuLauncher(this); break;
		case 2: item = ItemFactory::importIcon(this);          break;
		case 3: item = ItemFactory::importFileContent(this);   break;
	}
	restoreInsertionData();
	insertCreatedItem(item);
	resetInsertionData();
}

void Basket::insertColor(const QColor &color)
{
	Item *item = ItemFactory::createItemColor(color, this);
	restoreInsertionData();
	insertCreatedItem(item);
	resetInsertionData();
}

void Basket::pasteItem(QClipboard::Mode mode)
{
	if (redirectEditActions()) {
		if (m_editor->textEdit())
			m_editor->textEdit()->paste();
		else if (m_editor->lineEdit())
			m_editor->lineEdit()->paste();
	} else {
		Item *item = ItemFactory::dropItem(KApplication::clipboard()->data(mode/*QClipboard::Clipboard*/), this);
		if (item)
			insertCreatedItem(item);
	}
}

void Basket::insertCreatedItem(Item *item/*, bool edit*/)
{
	// Get the insertion data if the user clicked inside the basket:
	Item *clicked = m_clickedToInsert;
	int zone      = m_zoneToInsert;
	QPoint pos    = m_posToInsert;

	// If it isn't the case, use the default position:
	if (!clicked && (pos.x() < 0 || pos.y() < 0)) {
		if (isColumnsLayout()) {
			Item *column = (Settings::newItemsPlace == 0 ? firstItem() : lastItem());
			if (Settings::newItemsPlace == 0 && column->firstChild()) { // On Top, if at least one child in the column
				clicked = column->firstChild();
				zone    = Item::TopInsert;
			} else { // On Bottom
				clicked = column;
				zone    = Item::BottomColumn;
			}
		} else {
			pos = QPoint(0, 0);
		}
	}

	insertItem(item, clicked, zone, pos);
//	ensureVisibleItem(lastInsertedItem());
//	unselectAllBut(lastInsertedItem());
//	setFocusedItem(lastInsertedItem());
//	unselectAllBut(item);
	removeInserter(); // Case: user clicked below a column to insert, the item is inserted and doHoverEffects() put a new inserter below. We don't want it.
//	resetInsertTo();
	save();
///
///	if (edit)
///		itemEdit(item, /*justAdded=*/true);
}

void Basket::saveInsertionData()
{
	m_savedClickedToInsert = m_clickedToInsert;
	m_savedZoneToInsert    = m_zoneToInsert;
	m_savedPosToInsert     = m_posToInsert;
}

void Basket::restoreInsertionData()
{
	m_clickedToInsert = m_savedClickedToInsert;
	m_zoneToInsert    = m_savedZoneToInsert;
	m_posToInsert     = m_savedPosToInsert;
}

void Basket::resetInsertionData()
{
	m_clickedToInsert = 0;
	m_zoneToInsert    = 0;
	m_posToInsert     = QPoint(-1, -1);
}

void Basket::hideInsertPopupMenu()
{
	QTimer::singleShot( 50/*ms*/, this, SLOT(timeoutHideInsertPopupMenu()) );
}

void Basket::timeoutHideInsertPopupMenu()
{
	resetInsertionData();
}

void Basket::acceptDropEvent(QDropEvent *event, bool preCond)
{
	// FIXME: Should not accept all actions! Or not all actions (link not supported?!)
	event->acceptAction(preCond && 1);
	event->accept(preCond);
}










void Basket::contentsMouseReleaseEvent(QMouseEvent *event)
{
	// Now disallow drag:
	m_canDrag = false;

	// Cancel Resizer move:
	if (m_resizingItem) {
		m_resizingItem  = 0;
		m_pickedResizer = 0;
		m_lockedHovering = false;
		doHoverEffects();
		save();
	}

	// Cancel Item move:
/*	if (m_movingItem) {
		m_movingItem   = 0;
		m_pickedHandle = QPoint(0, 0);
		m_lockedHovering = false;
		//doHoverEffects();
		save();
	}
*/

	// Cancel Selection rectangle:
	if (m_isSelecting) {
		m_isSelecting = false;
		stopAutoScrollSelection();
		resetWasInLastSelectionRect();
		doHoverEffects();
		updateContents(m_selectionRect);
	}
	m_selectionStarted = false;

	// Do nothing if an action has already been made during mousePressEvent,
	// or if user made a selection and canceled it by regressing to a very small rectangle.
	if (m_noActionOnMouseRelease)
		return;
	// We immediatly set it to true, to avoid actions set on mouseRelease if NO mousePress event has been triggered.
	// This is the case when a popup menu is shown, and user click to the basket area to close it:
	// the menu then receive the mousePress event and the basket area ONLY receive the mouseRelease event.
	// Obviously, nothing should be done in this case:
	m_noActionOnMouseRelease = true;

	Item *clicked = itemAt(event->pos().x(), event->pos().y());
	if ( ! clicked ) {
		if (isFreeLayout() && event->button() == Qt::LeftButton) {
			clickedToInsert(event);
			save();
		}
		return;
	}
	Item::Zone zone = clicked->zoneAt( event->pos() - QPoint(clicked->x(), clicked->y()) );

	// Switch tag states:
	if (zone >= Item::Emblem0) {
		if (event->button() == Qt::LeftButton) {

			int icons = -1;
			for (State::List::iterator it = clicked->states().begin(); it != clicked->states().end(); ++it) {
				if ( ! (*it)->emblem().isEmpty() )
					icons++;
				if (icons == zone - Item::Emblem0) {
					State *state = (*it)->nextState();
					if (!state)
						return;
					it = clicked->states().insert(it, state);
					++it;
					clicked->states().remove(it);
					clicked->recomputeStyle();
					clicked->unbufferize();
					updateItem(clicked);
					searchAgain();
					save();
					break;
				}
			}
			return;

		}/* else if (event->button() == Qt::RightButton) {
		popupEmblemMenu(clicked, zone - Item::Emblem0);
		return;
	}*/
	}

	// Insert item or past clipboard:
	QString  text;
//	Item *item;
	QString  link;
	//int zone = zone;
	if (event->button() == Qt::MidButton && zone == Item::Resizer)
		return; //zone = clicked->zoneAt( event->pos() - QPoint(clicked->x(), clicked->y()), true );
	if (event->button() == Qt::RightButton && (clicked->isColumn() || zone == Item::Resizer))
		return;
	if (clicked->isGroup() && zone == Item::None)
		return;
	switch (zone) {
		case Item::Handle:
		case Item::Group:
			// We select item on mousePress if it was unselected or Ctrl is pressed.
			// But the user can want to drag select_s_ items, so it the item is selected, we only select it alone on mouseRelease:
			if (zone == Item::Handle || zone == Item::Group)
				if ( !(event->stateAfter() & Qt::ControlButton) && clicked->allSelected())
					unselectAllBut(clicked);
			//if (event->stateAfter() & Qt::ControlButton)
			//	clicked->setSelectedRecursivly(!clicked->isSelected());
			//else
			//	unselectAllBut(clicked);
			break;

		case Item::Custom0:
			unselectAllBut(clicked);
			itemOpen(clicked);
			break;

		case Item::GroupExpander: KMessageBox::information(viewport(), "Action not yet implemented"); break;

		case Item::TagsArrow:
//			if (!clicked->isSelected())
//				unselectAllBut(clicked);
//			popupTagsMenu(clicked);
			break;

		case Item::Link:
			link = clicked->linkAt(event->pos() - QPoint(clicked->x(), clicked->y()));
			if ( ! link.isEmpty() ) {
				KRun *run = new KRun(link); //  open the URL.
				run->setAutoDelete(true);
				break;
			} // If there is no link, edit item content
		case Item::Content:
			unselectAllBut(clicked);
			itemEdit(clicked, /*justAdded=*/false, event->pos());
			break;

		case Item::TopInsert:
		case Item::TopGroup:
		case Item::BottomInsert:
		case Item::BottomGroup:
		case Item::BottomColumn:
			clickedToInsert(event, clicked, zone);
			save();
			break;

		case Item::None:
		default:
			KMessageBox::information(viewport(),
					i18n("This message should never appears. If it's the case, this program is buggy!"
					"Please report the bug to the developer."));
			break;
	}
}

void Basket::contentsMouseDoubleClickEvent(QMouseEvent *event)
{
	Item *clicked = itemAt(event->pos().x(), event->pos().y());
	Item::Zone zone = (clicked ? clicked->zoneAt( event->pos() - QPoint(clicked->x(), clicked->y()) ) : Item::None);

	if (event->button() == Qt::LeftButton && (zone == Item::Group || zone == Item::Handle))
		doCopy(CopyToSelection);
	else
		contentsMousePressEvent(event);
}

void Basket::contentsMouseMoveEvent(QMouseEvent *event)
{
	// Drag the items:
	if (m_canDrag && (m_pressPos - event->/*globalPos*/pos()).manhattanLength() > KApplication::startDragDistance()) {
		m_canDrag          = false;
		m_isSelecting      = false; // Don't draw selection rectangle ater drag!
		m_selectionStarted = false;
		ItemSelection *selection = selectedItems();
		if (selection->firstStacked()) {
			QDragObject *d = ItemDrag::dragObject(selection, /*cutting=*/false, /*source=*/this); // d will be deleted by QT
		/*bool shouldRemove = */d->drag();
//		delete selection;

		// Never delete because URL is dragged and the file must be available for the extern appliation
//		if (shouldRemove && d->target() == 0) // If target is another application that request to remove the item
//			emit wantDelete(this);
		}
		return;
	}

	// Moving a Resizer:
	if (m_resizingItem) {
		int groupWidth = event->pos().x() - m_resizingItem->x() - m_pickedResizer;
		int minRight   = m_resizingItem->minRight();
		int maxRight   = 100 * contentsWidth(); // A big enough value (+infinity) for free layouts.
		Item *nextColumn = m_resizingItem->next();
		if (m_resizingItem->isColumn()) {
			if (nextColumn)
				maxRight = nextColumn->x() + nextColumn->rightLimit() - nextColumn->minRight() - Item::RESIZER_WIDTH;
			else
				maxRight = contentsWidth();
		}
		if (groupWidth > maxRight - m_resizingItem->x())
			groupWidth = maxRight - m_resizingItem->x();
		if (groupWidth < minRight - m_resizingItem->x())
			groupWidth = minRight - m_resizingItem->x();
		int delta = groupWidth - m_resizingItem->groupWidth();
		m_resizingItem->setGroupWidth(groupWidth);
		// If resizing columns:
		if (m_resizingItem->isColumn()) {
			Item *column = m_resizingItem;
			if ( (column = column->next()) ) {
				// Next columns should not have them X coordinate animated, because it would flicker:
				column->setXRecursivly(column->x() + delta);
				// And the resizer should resize the TWO sibling columns, and not push the other columns on th right:
				column->setGroupWidth(column->groupWidth() - delta);
			}
		}
		relayoutItems(true);
	}

	// Moving an Item:
/*	if (m_movingItem) {
		int x = event->pos().x() - m_pickedHandle.x();
		int y = event->pos().y() - m_pickedHandle.y();
		if (x < 0) x = 0;
		if (y < 0) y = 0;
		m_movingItem->setX(x);
		m_movingItem->setY(y);
		m_movingItem->relayoutAt(x, y, / *animate=* /false);
		relayoutItems(true);
	}
*/

	// Dragging the selection rectangle:
	if (m_selectionStarted)
		doAutoScrollSelection();

	doHoverEffects(event->pos());
}

void Basket::doAutoScrollSelection()
{
	static const int AUTO_SCROLL_MARGIN = 50;  // pixels
	static const int AUTO_SCROLL_DELAY  = 100; // milliseconds

	QPoint pos = viewport()->mapFromGlobal(QCursor::pos());

	// Do the selection:

	if (m_isSelecting)
		updateContents(m_selectionRect);

	m_selectionEndPoint = viewportToContents(pos);
	m_selectionRect = QRect(m_selectionBeginPoint, m_selectionEndPoint).normalize();
	if (m_selectionRect.left() < 0)                    m_selectionRect.setLeft(0);
	if (m_selectionRect.top() < 0)                     m_selectionRect.setTop(0);
	if (m_selectionRect.right() >= contentsWidth())    m_selectionRect.setRight(contentsWidth() - 1);
	if (m_selectionRect.bottom() >= contentsHeight())  m_selectionRect.setBottom(contentsHeight() - 1);

	if ( (m_selectionBeginPoint - m_selectionEndPoint).manhattanLength() > QApplication::startDragDistance() ) {
		m_isSelecting = true;
		selectItemsIn(m_selectionRect, m_selectionInvert);
		updateContents(m_selectionRect);
		m_noActionOnMouseRelease = true;
	} else {
			// If the user was selecting but cancel by making the rectangle too small, cancel it really!!!
		if (m_isSelecting) {
			if (m_selectionInvert)
				selectItemsIn(QRect(), m_selectionInvert);
			else
				unselectAllBut(0); // TODO: unselectAll();
		}
		if (m_isSelecting)
			resetWasInLastSelectionRect();
		m_isSelecting = false;
		stopAutoScrollSelection();
		return;
	}

	// Do the auto-scrolling:
	// FIXME: It's still flickering

	QRect insideRect(AUTO_SCROLL_MARGIN, AUTO_SCROLL_MARGIN, visibleWidth() - 2*AUTO_SCROLL_MARGIN, visibleHeight() - 2*AUTO_SCROLL_MARGIN);

	int dx = 0;
	int dy = 0;

	if (pos.y() < AUTO_SCROLL_MARGIN)
		dy = pos.y() - AUTO_SCROLL_MARGIN;
	else if (pos.y() > visibleHeight() - AUTO_SCROLL_MARGIN)
		dy = pos.y() - visibleHeight() + AUTO_SCROLL_MARGIN;

	if (pos.x() < AUTO_SCROLL_MARGIN)
		dx = pos.x() - AUTO_SCROLL_MARGIN;
	else if (pos.x() > visibleWidth() - AUTO_SCROLL_MARGIN)
		dx = pos.x() - visibleWidth() + AUTO_SCROLL_MARGIN;

	if (dx || dy) {
		kapp->sendPostedEvents(); // Do the repaints, because the scrolling will make the area to repaint to be wrong
		scrollBy(dx, dy);
		if (!m_autoScrollSelectionTimer.isActive())
			m_autoScrollSelectionTimer.start(AUTO_SCROLL_DELAY);
	} else
		stopAutoScrollSelection();
}

void Basket::stopAutoScrollSelection()
{
	m_autoScrollSelectionTimer.stop();
}

void Basket::resetWasInLastSelectionRect()
{
	Item *item = m_firstItem;
	while (item) {
		item->resetWasInLastSelectionRect();
		item = item->next();
	}
}

void Basket::selectAll()
{
	if (redirectEditActions()) {
		if (m_editor->textEdit())
			m_editor->textEdit()->selectAll();
		else if (m_editor->lineEdit())
			m_editor->lineEdit()->selectAll();
	} else {
		// TODO: First select all in the group, then in the parent group...
		FOR_EACH_ITEM (item)
			item->setSelectedRecursivly(true);
	}
}

void Basket::unselectAll()
{
	if (redirectEditActions()) {
		if (m_editor->textEdit()) {
			m_editor->textEdit()->removeSelection();
			selectionChangedInEditor(); // THIS IS NOT EMITED BY Qt!!!
		} else if (m_editor->lineEdit())
			m_editor->lineEdit()->deselect();
	} else {
		if (countSelecteds() > 0) // Optimisation
			FOR_EACH_ITEM (item)
				item->setSelectedRecursivly(false);
	}
}

void Basket::invertSelection()
{
	FOR_EACH_ITEM (item)
		item->invertSelectionRecursivly();
}

void Basket::unselectAllBut(Item *toSelect)
{
	FOR_EACH_ITEM (item)
		item->unselectAllBut(toSelect);
/*	Item *item = m_firstItem;
	while (item) {
		item->unselectAllBut(toSelect);
		item = item->next();
	}*/
}

void Basket::invertSelectionOf(Item *toSelect)
{
	Item *item = m_firstItem;
	while (item) {
		item->invertSelectionOf(toSelect);
		item = item->next();
	}
}

void Basket::selectItemsIn(const QRect &rect, bool invertSelection)
{
	Item *item = m_firstItem;
	while (item) {
		item->selectIn(rect, invertSelection);
		item = item->next();
	}
}

void Basket::doHoverEffects()
{
	doHoverEffects(  viewportToContents( viewport()->mapFromGlobal(QCursor::pos()) )  );
}

void Basket::doHoverEffects(Item *item, Item::Zone zone, const QPoint &pos)
{
	// Inform the old and new hovered item (if any):
	Item *oldHoveredItem = m_hoveredItem;
	if (item != m_hoveredItem) {
		if (m_hoveredItem) {
			m_hoveredItem->setHovered(false);
			m_hoveredItem->setHoveredZone(Item::None);
			updateItem(m_hoveredItem);
		}
		m_hoveredItem = item;
		if (item)
			item->setHovered(true);
	}

	// If we are hovering an item, compute which zone is hovered and inform the item:
	if (m_hoveredItem) {
		if (zone != m_hoveredZone || oldHoveredItem != m_hoveredItem) {
			m_hoveredZone = zone;
			m_hoveredItem->setCursor(zone);
			updateItem(m_hoveredItem);
		}
		m_hoveredItem->setHoveredZone(zone);
		// If we are hovering an insert line zone, update this thing:
		if (zone == Item::TopInsert || zone == Item::TopGroup || zone == Item::BottomInsert || zone == Item::BottomGroup || zone == Item::BottomColumn)
			placeInserter(m_hoveredItem, zone);
		else
			removeInserter();
		// If we are hovering an embedded link in a rich text element, show the destination in the statusbar:
		if (zone == Item::Link)
			emit setStatusBarText(m_hoveredItem->linkAt( pos - QPoint(m_hoveredItem->x(), m_hoveredItem->y()) ));
		else if (m_hoveredItem->content())
			emit setStatusBarText(m_hoveredItem->content()->statusBarMessage(m_hoveredZone));//resetStatusBarText();
	// If we aren't hovering an item, reset all:
	} else {
		if (isFreeLayout() && !isSelecting())
			viewport()->setCursor(Qt::CrossCursor);
		else
			viewport()->unsetCursor();
		m_hoveredZone = Item::None;
		removeInserter();
		emit resetStatusBarText();
	}
}

void Basket::doHoverEffects(const QPoint &pos)
{
	if (isDuringEdit())
		viewport()->unsetCursor();

	// Do we have the right to do hover effects?
	if ( ! m_loaded || m_lockedHovering)
		return;

	// enterEvent() (mouse enter in the widget) set m_underMouse to true, and leaveEvent() make it false.
	// But some times the enterEvent() is not trigerred: eg. when dragging the scrollbar:
	// Ending the drag INSIDE the basket area will make NO hoverEffects() because m_underMouse is false.
	// User need to leave the area and re-enter it to get effects.
	// This hack solve that by dismissing the m_underMouse variable:
	bool underMouse = Global::mainContainer->currentBasket() == this && QRect(contentsX(), contentsY(), visibleWidth(), visibleHeight()).contains(pos);

	// Don't do hover effects when a popup menu is opened.
	// Primarily because the basket area will only receive mouseEnterEvent and mouveLeaveEvent.
	// It willn't be noticed of mouseMoveEvent, which would result in a apparently broken application state:
	if (kapp->activePopupWidget())
		underMouse = false;

	// Compute which item is hovered:
	Item       *item = (m_isSelecting || !underMouse ? 0 : itemAt(pos.x(), pos.y()));
	Item::Zone  zone = (item ? item->zoneAt( pos - QPoint(item->x(), item->y()), isDuringDrag() ) : Item::None);

	// Inform the old and new hovered item (if any) and update the areas:
	doHoverEffects(item, zone, pos);
}

void Basket::removeInserter()
{
	if (m_inserterShown) { // Do not hide (and then update/repaint the view) if it is already hidden!
		m_inserterShown = false;
		updateContents(m_inserterRect);
	}
}

void Basket::placeInserter(Item *item, int zone)
{
	// Remove the inserter:
	if (!item) {
		removeInserter();
		return;
	}

	// Update the old position:
	if (inserterShown())
		updateContents(m_inserterRect);
	// Some comodities:
	m_inserterShown = true;
	m_inserterTop   = (zone == Item::TopGroup || zone == Item::TopInsert);
	m_inserterGroup = (zone == Item::TopGroup || zone == Item::BottomGroup);
	// X and width:
	int groupIndent = (item->isGroup() ? item->width() : Item::HANDLE_WIDTH);
	int x     = item->x();
	int width = (item->isGroup() ? item->rightLimit() - item->x() : item->width());
	if (m_inserterGroup) {
		x     += groupIndent;
		width -= groupIndent;
	}
	m_inserterSplit = (item && !item->isGroup() && !item->isFree() && !item->isColumn());
//	if (item->isGroup())
//		width = item->rightLimit() - item->x() - (m_inserterGroup ? groupIndent : 0);
	// Y:
	int y = item->y() - (m_inserterGroup && m_inserterTop ? 1 : 3);
	if (!m_inserterTop)
		y += (item->isColumn() ? item->finalHeight() : item->height());
	// Assigning result:
	m_inserterRect = QRect(x, y, width, 6 - (m_inserterGroup ? 2 : 0));
	// Update the new position:
	updateContents(m_inserterRect);
}

inline void drawLineByRect(QPainter &painter, int x, int y, int width, int height)
{
	painter.drawLine(x, y, x + width - 1, y + height - 1);
}

void Basket::drawInserter(QPainter &painter, int xPainter, int yPainter/*, int widthPainter, int heightPainter*/)
{
	if (!m_inserterShown)
		return;

	QRect rect = m_inserterRect; // For shorter code-lines when drawing!
	rect.moveBy(-xPainter, -yPainter);
	int lineY  = (m_inserterGroup && m_inserterTop ? 0 : 2);
	int roundY = (m_inserterGroup && m_inserterTop ? 0 : 1);

	QColor dark  = KApplication::palette().active().dark();
	QColor light = dark.light().light();
	if (m_inserterGroup) {
		//dark  = Utilities::mixColor(dark,  KGlobalSettings::highlightColor());
		light = Utilities::mixColor(light, KGlobalSettings::highlightColor());
	}
	painter.setPen(dark);
	// The horizontal line:
	//painter.drawRect(       rect.x(),                    rect.y() + lineY,  rect.width(), 2);
	int width = rect.width() - 4;
	drawGradient(&painter, dark,  light, rect.x() + 2,           rect.y() + lineY, width/2,         2, /*sunken=*/false, /*horz=*/false, /*flat=*/false);
	drawGradient(&painter, light, dark,  rect.x() + 2 + width/2, rect.y() + lineY, width - width/2, 2, /*sunken=*/false, /*horz=*/false, /*flat=*/false);
	// The left-most and right-most edges (biggest vertical lines):
	drawLineByRect(painter, rect.x(),                    rect.y(),          1,            (m_inserterGroup ? 4 : 6));
	drawLineByRect(painter, rect.x() + rect.width() - 1, rect.y(),          1,            (m_inserterGroup ? 4 : 6));
	// The left and right mid vertical lines:
	drawLineByRect(painter, rect.x() + 1,                rect.y() + roundY, 1,            (m_inserterGroup ? 3 : 4));
	drawLineByRect(painter, rect.x() + rect.width() - 2, rect.y() + roundY, 1,            (m_inserterGroup ? 3 : 4));
	// Draw the split as a feedback to know where is the limit between insert and group:
	if (m_inserterSplit) {
		int itemWidth = rect.width() + (m_inserterGroup ? Item::HANDLE_WIDTH : 0);
		int xSplit = rect.x() - (m_inserterGroup ? Item::HANDLE_WIDTH : 0) + itemWidth / 2;
		painter.setPen(Utilities::mixColor(dark, light));
		painter.drawRect(xSplit - 2, rect.y() + lineY, 4, 2);
		painter.setPen(dark);
		painter.drawRect(xSplit - 1, rect.y() + lineY, 2, 2);
	}
}

void Basket::maybeTip(const QPoint &pos)
{
	if ( !m_loaded || !Settings::showItemsToolTip() )
		return;

	QString message;
	QRect   rect;

	QPoint contentPos = viewportToContents(pos);
	Item *item = itemAt(contentPos.x(), contentPos.y());

	if (isDuringEdit()) {
		message = i18n("Validate item edition"); // TODO: i18n("Cancel item addition");    i18n("Validate item addition");   ""
		QValueList<QRect> areas;
		areas.clear();
		areas.append( QRect(0, 0, contentsWidth(), contentsHeight()) );
		QRect editorRect(m_editor->widget()->x(), m_editor->widget()->y(), m_editor->widget()->width(), m_editor->widget()->height());
		substractRectOnAreas(editorRect, areas, /*andRemove=*/true);
		for (QValueList<QRect>::iterator it = areas.begin(); it != areas.end(); ++it)
			if ((*it).contains(pos))
				rect = *it;
	} else if (!item && isFreeLayout()) {
		message = i18n("Insert an item here\nRight click for more options");
		QRect itRect;
		for (QValueList<QRect>::iterator it = m_blankAreas.begin(); it != m_blankAreas.end(); ++it) {
			itRect = QRect(0, 0, visibleWidth(), visibleHeight()).intersect(*it);
			if (itRect.contains(contentPos)) {
				rect = itRect;
				rect.moveLeft(rect.left() - contentsX());
				rect.moveTop( rect.top()  - contentsY());
				break;
			}
		}
	} else {
		if (!item)
			return;

		Item::Zone zone = item->zoneAt( contentPos - QPoint(item->x(), item->y()) );
		switch (zone) {
			case Item::Resizer:       message = (item->isColumn() ?
			                                    i18n("Resize those columns") :
			                                    (item->isGroup() ?
			                                    i18n("Resize this group") :
			                                    i18n("Resize this item")));                 break;
			case Item::Handle:        message = i18n("Select or move this item");           break;
			case Item::Group:         message = i18n("Select or move this group");          break;
			case Item::TagsArrow:     message = i18n("Assign or remove tags to this item");
			                          if (item->states().count() > 0) {
			                          	message = "<qt><nobr>" + message + "</nobr><br>" + i18n("<b>Assigned Tags</b>: %1");
			                          	QString tagsString = "";
			                          	for (State::List::iterator it = item->states().begin(); it != item->states().end(); ++it) {
			                          		QString tagName = "<nobr>" + Utilities::textToHTMLWithoutP((*it)->fullName()) + "</nobr>";
			                          		if (tagsString.isEmpty())
			                          			tagsString = tagName;
			                          		else
			                          			tagsString = i18n("%1, %2").arg(tagsString, tagName);
			                          	}
			                          	message = message.arg(tagsString);
			                          }
			                          break;
			case Item::Custom0:       message = item->content()->zoneTip(zone);             break; //"Open this link/Open this file/Open this sound file/Launch this application"
			case Item::GroupExpander: message = (item->isFolded() ?
			                                    i18n("Expand this group") :
			                                    i18n("Collapse this group"));               break;
			case Item::Link:
			case Item::Content:       message = item->content()->editToolTipText();         break;
			case Item::TopInsert:
			case Item::BottomInsert:  message = i18n("Insert an item here\nRight click for more options");             break;
			case Item::TopGroup:      message = i18n("Group an item to the one below\nRight click for more options");  break;
			case Item::BottomGroup:   message = i18n("Group an item to the one above\nRight click for more options");  break;
			case Item::BottomColumn:  message = i18n("Insert an item here\nRight click for more options");             break;
			case Item::None:          message = "** Zone NONE: internal error **";          break;
			default:
				if (zone >= Item::Emblem0)
					message = item->stateForEmblemNumber(zone - Item::Emblem0)->fullName();
				else
					message = "";
				break;
		}

		if (zone == Item::Content || zone == Item::Link) {
			QStringList keys;
			QStringList values;

			item->content()->toolTipInfos(&keys, &values);
			keys.append(i18n("Added"));
			keys.append(i18n("Last Modification"));
			values.append(item->addedStringDate());
			values.append(item->lastModificationStringDate());

			message = "<qt><nobr>" + message;
			QStringList::iterator key;
			QStringList::iterator value;
			for (key = keys.begin(), value = values.begin(); key != keys.end() && value != values.end(); ++key, ++value)
				message += "<br>" + i18n("of the form 'key: value'", "<b>%1</b>: %2").arg(*key, *value);
			message += "</nobr></qt>";
		} else if (m_inserterSplit && (zone == Item::TopInsert || zone == Item::BottomInsert))
			message += "\n" + i18n("Click on the right to group instead of insert");
		else if (m_inserterSplit && (zone == Item::TopGroup || zone == Item::BottomGroup))
			message += "\n" + i18n("Click on the left to insert instead of group");

		rect = item->zoneRect( zone, contentPos - QPoint(item->x(), item->y()) );

		rect.moveLeft(rect.left() - contentsX());
		rect.moveTop( rect.top()  - contentsY());

		// DEBUG:
		/*	QRect r = rect;
		r.moveLeft(r.left() + item->x());
		r.moveTop( r.top()  + item->y());
		QPainter painter(viewport());
		painter.drawRect(r);
		painter.end();*/

		rect.moveLeft(rect.left() + item->x());
		rect.moveTop( rect.top()  + item->y());
	}

	tip(rect, message/*, QRect(item->x(), item->y() + item->height(), 1024, 20)*/);
}

Item* Basket::lastItem()
{
	Item *item = firstItem();
	while (item && item->next())
		item = item->next();
	return item;
}

Item* Basket::itemAt(int x, int y)
{
	if (x < 0 || x > contentsWidth() || y < 0 || y > contentsHeight())
		return NULL;

	// When resizing an item/group, keep it highlighted:
	if (m_resizingItem)
		return m_resizingItem;

	// Search and return the hovered item:
	Item *item = m_firstItem;
	Item *possibleItem;
	while (item) {
		possibleItem = item->itemAt(x, y);
//		if (possibleItem)
//			return possibleItem;
		if (possibleItem) {
			if (draggedItems().contains(possibleItem))
				return 0;
			else
				return possibleItem;
		}
		item = item->next();
	}

	// If the basket is layouted in columns, return one of the columns to be able to add items in them:
	if (isColumnsLayout()) {
		Item *column = m_firstItem;
		while (column) {
			if (x >= column->x() && x < column->rightLimit())
				return column;
			column = column->next();
		}
	}

	// Nothing found, no item is hovered:
	return NULL;
}

Basket::~Basket()
{
	delete m_action;
	// TODO: Delete Items (and then ItemContents)!
}

void Basket::viewportResizeEvent(QResizeEvent *)
{
	relayoutItems(true);
//	placeEditor();
	doHoverEffects();
}

void Basket::animateLoad()
{
	const int viewHeight = contentsY() + visibleHeight();

	QTime t = QTime::currentTime(); // Set random seed
	srand(t.hour()*12 + t.minute()*60 + t.second()*60);

	Item *item = firstItem();
	while (item) {
		if ((item->finalY() < viewHeight) && item->matching())
			item->initAnimationLoad();
		item = item->next();
	}

	m_loaded = true;
}

QColor Basket::selectionRectInsideColor()
{
	return Utilities::mixColor(Utilities::mixColor(backgroundColor(), KGlobalSettings::highlightColor()), backgroundColor());
}


QColor alphaBlendColors(const QColor &bgColor, const QColor &fgColor, const int a)
{
	// normal button...
	QRgb rgb = bgColor.rgb();
	QRgb rgb_b = fgColor.rgb();
	int alpha = a;
	if (alpha>255) alpha = 255;
	if (alpha<0) alpha = 0;
	int inv_alpha = 255 - alpha;
	QColor result = QColor( qRgb(qRed(rgb_b)*inv_alpha/255 + qRed(rgb)*alpha/255,
	                        qGreen(rgb_b)*inv_alpha/255 + qGreen(rgb)*alpha/255,
	                        qBlue(rgb_b)*inv_alpha/255 + qBlue(rgb)*alpha/255) );

	return result;
}


void Basket::drawContents(QPainter *painter, int clipX, int clipY, int clipWidth, int clipHeight)
{
	// Start the load the first time the basket is shown:
	if (!m_loadingLaunched)
		QTimer::singleShot( 0, this, SLOT(load()) );

	QBrush brush(backgroundColor()); // FIXME: share it for all the basket?
	QRect clipRect(clipX, clipY, clipWidth, clipHeight);

	// Draw items (but not when it's not loaded yet):
	Item *item = (m_loaded ? m_firstItem : 0);
	while (item) {
		item->draw(painter, clipRect);
		item = item->next();
	}

	// Draw loading message:
	if (!m_loaded) {
		QPixmap pixmap(visibleWidth(), visibleHeight()); // TODO: Clip it to asked size only!
		QPainter painter2(&pixmap);
		QSimpleRichText rt(QString("<center>%1</center>").arg(i18n("Loading...")), QScrollView::font());
		rt.setWidth(visibleWidth());
		int hrt = rt.height();
		painter2.fillRect(0, 0, visibleWidth(), visibleHeight(), brush);
		blendBackground(painter2, QRect(0, 0, visibleWidth(), visibleHeight()), -1, -1, /*opaque=*/true);
		// The "Loading..." text is not shown immediatly but after 1/2 second.
		// After that, the text slowly appear in about 1/2 second.
		// A timer increment m_loadingFrames every 50 ms. When it's 10, the text can be shown:
/*		static const int showDelay = 0/ *ms* /;
		static const int fadeDelay = 500/ *ms* /;
		static const int frameRate = 50/ *ms* /;
		if (m_loadingFrames >= showDelay / frameRate) {
			std::cout << "repainting..." << std::endl;
			QColor theTextColor = textColor();
			if (m_loadingFrames < (showDelay + fadeDelay) / frameRate)
				theTextColor = alphaBlendColors(backgroundColor(), textColor(), (m_loadingFrames-showDelay/frameRate) * 255 / (fadeDelay/frameRate));
			std::cout << theTextColor.name() << std::endl;
			QColorGroup cg = colorGroup();
			cg.setColor(QColorGroup::Text, theTextColor);
			rt.draw(&painter2, 0, (visibleHeight() - hrt) / 2, QRect(), cg);
		}*/
		QColorGroup cg = colorGroup();
		cg.setColor(QColorGroup::Text, textColor());
		rt.draw(&painter2, 0, (visibleHeight() - hrt) / 2, QRect(), cg);
		painter2.end();
		painter->drawPixmap(0, 0, pixmap);
		return; // TODO: Animate the hidding of the "Loading..." message... // TODO: Also clip to the wanted rectangle
	}

	// We will draw blank areas below.
	// For each rectangle to be draw there is three ways to do so:
	// - The rectangle is full of the background COLOR  => we fill a rect directly on screen
	// - The rectangle is full of the background PIXMAP => we draw it directly on screen (we draw m_opaqueBackgroundPixmap that is not transparent)
	// - The rectangle contains the resizer             => We draw it on an offscreen buffer and then paint the buffer on screen
	// If the background image is not tiled, we know that recomputeBlankRects() broken rects so that they are full of either background pixmap or color, but not a mix.

	// Draw blank areas (see the last preparation above):
	QPixmap  pixmap;
	QPainter painter2;
	QRect    rect;
	for (QValueList<QRect>::iterator it = m_blankAreas.begin(); it != m_blankAreas.end(); ++it) {
		rect = clipRect.intersect(*it);
		if (rect.width() > 0 && rect.height() > 0) {
			// If there is an inserter to draw, draw the image off screen,
			// apply the inserter and then draw the image on screen:
			if (  (inserterShown() && rect.intersects(inserterRect()))  ||  (m_isSelecting && rect.intersects(m_selectionRect))  ) {
				pixmap.resize(rect.width(), rect.height());
				painter2.begin(&pixmap);
				painter2.fillRect(0, 0, rect.width(), rect.height(), backgroundColor());
				blendBackground(painter2, rect, -1, -1, /*opaque=*/true);
				// Draw inserter:
				if (inserterShown() && rect.intersects(inserterRect()))
					drawInserter(painter2, rect.x(), rect.y());
				// Draw selection rect:
				if (m_isSelecting && rect.intersects(m_selectionRect)) {
					QRect selectionRect = m_selectionRect;
					selectionRect.moveBy(-rect.x(), -rect.y());
					QRect selectionRectInside(selectionRect.x() + 1, selectionRect.y() + 1, selectionRect.width() - 2, selectionRect.height() - 2);
					if (selectionRectInside.width() > 0 && selectionRectInside.height() > 0) {
						QColor insideColor = selectionRectInsideColor();
						painter2.fillRect(selectionRectInside, insideColor);
						selectionRectInside.moveBy(rect.x(), rect.y());
						blendBackground(painter2, selectionRectInside, rect.x(), rect.y(), true, /*&*/m_selectedBackgroundPixmap);
					}
					painter2.setPen(KGlobalSettings::highlightColor().dark());
					painter2.drawRect(selectionRect);
					painter2.setPen(Utilities::mixColor(KGlobalSettings::highlightColor().dark(), backgroundColor()));
					painter2.drawPoint(selectionRect.topLeft());
					painter2.drawPoint(selectionRect.topRight());
					painter2.drawPoint(selectionRect.bottomLeft());
					painter2.drawPoint(selectionRect.bottomRight());
				}
				painter2.end();
				painter->drawPixmap(rect.x(), rect.y(), pixmap);
			// If it's only a blank rectangle to draw, draw it directly on screen (faster!!!):
			} else if ( ! hasBackgroundImage() ) {
				painter->fillRect(rect, backgroundColor());
			// It's either a background pixmap to draw or a background color to fill:
			} else {
				if (isTiledBackground() || (rect.x() < backgroundPixmap()->width() && rect.y() < backgroundPixmap()->height()))
					blendBackground(*painter, rect, 0, 0, /*opaque=*/true);
				else
					painter->fillRect(rect, backgroundColor());
			}
		}
	}

/* DEBUG:
	static int colorIndex = 0;
	++colorIndex;
	QColor colors[] = { Qt::red, Qt::darkRed, Qt::green, Qt::darkGreen, Qt::blue, Qt::darkBlue, Qt::cyan, Qt::darkCyan, Qt::magenta, Qt::darkMagenta };
	painter->setPen(colors[colorIndex % 10]);
	painter->drawRect(clipRect);
*/
}

/*  rect(x,y,width,height)==(xBackgroundToDraw,yBackgroundToDraw,widthToDraw,heightToDraw)
 */
void Basket::blendBackground(QPainter &painter, const QRect &rect, int xPainter, int yPainter, bool opaque, QPixmap *bg)
{
	if (xPainter == -1 && yPainter == -1) {
		xPainter = rect.x();
		yPainter = rect.y();
	}

	if (hasBackgroundImage()) {
		const QPixmap *bgPixmap = (bg ? /* * */bg : (opaque ? m_opaqueBackgroundPixmap : m_backgroundPixmap));
		if (isTiledBackground())
			painter.drawTiledPixmap(rect.x() - xPainter, rect.y() - yPainter, rect.width(), rect.height(), *bgPixmap, rect.x(), rect.y());
		else
			painter.drawPixmap(rect.x() - xPainter, rect.y() - yPainter, *bgPixmap, rect.x(), rect.y(), rect.width(), rect.height());
	}
}

void Basket::recomputeBlankRects()
{
	m_blankAreas.clear();
	m_blankAreas.append( QRect(0, 0, contentsWidth(), contentsHeight()) );

	Item *item = m_firstItem;
	while (item) {
		item->recomputeBlankRects(m_blankAreas);
		item = item->next();
	}

	// See the drawing of blank areas in Basket::drawContents()
	if (hasBackgroundImage() && ! isTiledBackground())
		substractRectOnAreas( QRect(0, 0, backgroundPixmap()->width(), backgroundPixmap()->height()), m_blankAreas, false );
}

void Basket::addAnimatedItem(Item *item)
{
	if (m_animatedItems.isEmpty()) {
		m_animationTimer.start(FRAME_DELAY);
		m_lastFrameTime = QTime::currentTime();
	}

	m_animatedItems.append(item);
}

void Basket::unsetItemsWidth()
{
	Item *item = m_firstItem;
	while (item) {
		item->unsetWidth();
		item = item->next();
	}
}

void Basket::relayoutItems(bool animate)
{
	if (!Settings::playAnimations())
		animate = false;

	if (!animate) {
		m_animatedItems.clear();
		m_animationTimer.stop();
	}

	int h     = 0;
	tmpWidth  = 0;
	tmpHeight = 0;
	Item *item = m_firstItem;
	while (item) {
		if (item->matching()) {
			item->relayoutAt(0, h, animate);
			if (item->isColumn() && item->hasResizer()) {
				int minGroupWidth = item->minRight() - item->x();
				if (item->groupWidth() < minGroupWidth) {
					item->setGroupWidth(minGroupWidth);
					relayoutItems(animate); // Redo the thing, but this time it should not recurse
					return;
				}
			}
			h += item->finalHeight();
		}
		item = item->next();
	}

	resizeContents( QMAX(tmpWidth, visibleWidth()), QMAX(tmpHeight, visibleHeight()) );
	recomputeBlankRects();
	placeEditor();//
	updateContents();
}

void Basket::updateItem(Item *item)
{
	updateContents(item->rect());
	if (item->hasResizer())
		updateContents(item->resizerRect());
}

void Basket::animateObjects()
{
	QValueList<Item*>::iterator it;
	for (it = m_animatedItems.begin(); it != m_animatedItems.end(); )
//		if ((*it)->y() >= contentsY() && (*it)->bottom() <= contentsY() + contentsWidth())
//			updateItem(*it);
		if ((*it)->advance())
			it = m_animatedItems.remove(it);
	else {
//			if ((*it)->y() >= contentsY() && (*it)->bottom() <= contentsY() + contentsWidth())
//				updateItem(*it);
		++it;
	}

	if (m_animatedItems.isEmpty()) {
		// Stop animation timer:
		m_animationTimer.stop();
		// Reset all onTop items:
		Item* item = m_firstItem;
		while (item) {
			item->setOnTop(false);
			item = item->next();
		}
	}

	// We refresh content if it was the last frame,
	// or if the drawing of the last frame was not too long.
	if (!m_animationTimer.isActive() || (m_lastFrameTime.msecsTo(QTime::currentTime()) < FRAME_DELAY*11/10)) { // *11/10 == *1.1 : We keep a 0.1 security margin
		m_lastFrameTime = m_lastFrameTime.addMSecs(FRAME_DELAY);                                               // because timers are not accurate and can trigger late
		//m_lastFrameTime = QTime::currentTime();
//std::cout << ">>" << m_lastFrameTime.toString("hh:mm:ss.zzz") << std::endl;
		if (m_underMouse)
			doHoverEffects();
		recomputeBlankRects();
		//relayoutItems(true); // In case an animated item was to the contents view boundaries, resize the view!
		updateContents();
	// If the drawing of the last frame was too long, we skip the drawing of the current and do the next one:
	} else {
		m_lastFrameTime = m_lastFrameTime.addMSecs(FRAME_DELAY);
//std::cout << "+=" << m_lastFrameTime.toString("hh:mm:ss.zzz") << std::endl;
		animateObjects();
	}

	doHoverEffects();
	placeEditor();

/*	int delta = m_deltaY / 3;
	if (delta == 0       && m_deltaY != 0)
	delta = (m_deltaY > 0 ? 1 : -1);
	m_deltaY -= delta;
	resizeContents(contentsWidth(), contentsHeight() + delta); //m_lastItem->y() + m_lastItem->height()
*/
}

void Basket::popupEmblemMenu(Item *item, int emblemNumber)
{
	m_tagPopupItem = item;
	State *state = item->stateForEmblemNumber(emblemNumber);
	Tag *tag = state->parentTag();
	m_tagPopup = tag;

	KPopupMenu menu(this);
	if (tag->countStates() == 1) {
		menu.insertTitle(/*SmallIcon(state->icon()), */tag->name());
		menu.insertItem( SmallIconSet("editdelete"), i18n("&Remove"),             1 );
		menu.insertItem( SmallIconSet("configure"),  i18n("&Customize..."),       2 );
		menu.insertSeparator();
		menu.insertItem( SmallIconSet("filter"),     i18n("&Filter on This Tag"), 3 );
	} else {
		menu.insertTitle(tag->name());
		QValueList<State*>::iterator it;
		State *currentState;
		int i = 10;
		int id;
		for (it = tag->states().begin(); it != tag->states().end(); ++it) {
			currentState = *it;
			id = menu.insertItem(StateMenuItem::radioButtonIconSet(state == currentState, menu.colorGroup()), new StateMenuItem(currentState, false), i );
			std::cout << "-> " << currentState->parentTag()->shortcut().toStringInternal() << std::endl;
			if (currentState->parentTag() && !currentState->parentTag()->shortcut().isNull())
				menu.setAccel( currentState->parentTag()->shortcut().operator QKeySequence(), id );
			std::cout << currentState->parentTag()->shortcut().toStringInternal() << std::endl;
//			if (currentState->icon().isEmpty())
//				id = menu.insertItem(                                     new StateMenuItem(currentState, state == currentState), i );
//			else
//				id = menu.insertItem( SmallIconSet(currentState->icon()), new StateMenuItem(currentState, state == currentState), i );
//			menu.setItemChecked(id, state == currentState);
			++i;
		}
		menu.insertSeparator();
		menu.insertItem( new IndentedMenuItem(i18n("&Remove"),               "editdelete"), 1 );
		menu.insertItem( new IndentedMenuItem(i18n("&Customize..."),         "configure"),  2 );
		menu.insertSeparator();
		menu.insertItem( new IndentedMenuItem(i18n("&Filter on This Tag"),   "filter"),     3 );
		menu.insertItem( new IndentedMenuItem(i18n("Filter on This &State"), "filter"),     4 );
	}

	connect( &menu, SIGNAL(activated(int)), this, SLOT(toggledStateInMenu(int)) );
	connect( &menu, SIGNAL(aboutToHide()),  this, SLOT(unlockHovering())        );
	connect( &menu, SIGNAL(aboutToHide()),  this, SLOT(disableNextClick())      );

	m_lockedHovering = true;
	menu.exec(QCursor::pos());
}

void Basket::toggledStateInMenu(int id)
{
	if (id == 1) {
		removeTagFromSelectedItems(m_tagPopup);
		//m_tagPopupItem->removeTag(m_tagPopup);
		//m_tagPopupItem->setWidth(0); // To force a new layout computation
		searchAgain();
		save();
		//relayoutItems(true);
		return;
	}
	if (id == 2) {
		KMessageBox::information(viewport(),
								 "<qt>This is not implemented yet.<br>In a future version, you will be able to add, remove or modify tags "
										 "by setting icon, states, background color, font, font size, bold, italic, underline, text color...<br>"
										 "But for now, I'm afraid you have to wait ;-)");
		return;
	}
	if (id == 3) { // Filter on This Tag
		decoration()->searchBar()->searchTag(m_tagPopup);
		return;
	}
	if (id == 4) { // Filter on This State
		decoration()->searchBar()->searchState(m_tagPopupItem->stateOfTag(m_tagPopup));
		return;
	}

	/*addStateToSelectedItems*/changeStateOfSelectedItems(m_tagPopup->states()[id - 10] /*, orReplace=true*/);
	//m_tagPopupItem->addState(m_tagPopup->states()[id - 10], true);
	searchAgain();
	save();
}

void Basket::popupTagsMenu(Item *item)
{
	m_tagPopupItem = item;

	KPopupMenu menu(this);
	menu.insertTitle(i18n("Tags"));
	QValueList<Tag*>::iterator it;
	Tag *currentTag;
	State *currentState;
	int i = 10;
	int id;
	for (it = Tag::all.begin(); it != Tag::all.end(); ++it) {
		// Current tag and first state of it:
		currentTag = *it;
		currentState = currentTag->states().first();
		id = menu.insertItem(StateMenuItem::checkBoxIconSet(item->hasTag(currentTag), menu.colorGroup()), new StateMenuItem(currentState, true), i );
	//	std::cout << "-> " << currentTag->shortcut().toStringInternal() << std::endl;
	//	QKeySequence key = currentTag->shortcut().operator QKeySequence();
	//	std::cout << "QKS:" << key.operator QString() << std::endl;
		//if (!currentTag->shortcut().isNull())
		menu.setAccel( QKeySequence(Qt::CTRL+Qt::Key_O)/*currentTag->shortcut().operator QKeySequence()*//*QKeySequence("Ctrl+O")*/, id );
	//	std::cout << currentTag->shortcut().toStringInternal() << std::endl;
//		if (currentState->icon().isEmpty())
//			id = menu.insertItem(                                     new StateMenuItem(currentState, item->hasTag(currentTag), true), i );
//		else
//			id = menu.insertItem( SmallIconSet(currentState->icon()), new StateMenuItem(currentState, item->hasTag(currentTag), true), i );
//		menu.setItemChecked(id, item->hasTag(currentTag));
		++i;
	}

	menu.insertSeparator();
//	menu.insertItem( /*SmallIconSet("editdelete"),*/ "&Assign new Tag...", 1 );
	//id = menu.insertItem( SmallIconSet("editdelete"), "&Remove All", -2 );
	//if (item->states().isEmpty())
	//	menu.setItemEnabled(id, false);
//	menu.insertItem( SmallIconSet("configure"),  "&Customize...", 3 );
	menu.insertItem(      new IndentedMenuItem(i18n("&Assign new Tag...")),          1 );
	id = menu.insertItem( new IndentedMenuItem(i18n("&Remove All"),   "editdelete"), 2 );
	menu.insertItem(      new IndentedMenuItem(i18n("&Customize..."), "configure"),  3 );

	if (!selectedItemsHaveTags())//item->states().isEmpty())
		menu.setItemEnabled(id, false);

	connect( &menu, SIGNAL(activated(int)), this, SLOT(toggledTagInMenu(int)) );
	connect( &menu, SIGNAL(aboutToHide()),  this, SLOT(unlockHovering())      );
	connect( &menu, SIGNAL(aboutToHide()),  this, SLOT(disableNextClick())    );

	m_lockedHovering = true;
	menu.exec(QCursor::pos());
}

void Basket::unlockHovering()
{
	m_lockedHovering = false;
	doHoverEffects();
}

void Basket::toggledTagInMenu(int id)
{
	if (id == 1) {
		KMessageBox::information(viewport(),
								 "<qt>Sorry... Not implemented yet! Please come back later :-)<br>"
										 "This thing will allow to add a new tag and assign it to the item in one go.");
		return;
	}
	if (id == 2) {
		removeAllTagsFromSelectedItems();//m_tagPopupItem->removeAllTags();
		//m_tagPopupItem->setWidth(0); // To force a new layout computation
		searchAgain();
		save();
		//relayoutItems(true);
		return;
	}
	if (id == 3) {
		KMessageBox::information(viewport(),
								 "<qt>This is not implemented yet.<br>In a future version, you will be able to add, remove or modify tags "
										 "by setting icon, states, background color, font, font size, bold, italic, underline, text color...<br>"
										 "But for now, I'm afraid you have to wait ;-)");
		return;
	}

	Tag *tag = Tag::all[id - 10];
	if (!tag)
		return;

	if (m_tagPopupItem->hasTag(tag))
		removeTagFromSelectedItems(tag);//m_tagPopupItem->removeTag(tag);
	else
		addTagToSelectedItems(tag);//m_tagPopupItem->addTag(tag);
	m_tagPopupItem->setWidth(0); // To force a new layout computation
	searchAgain();
	save();
	//relayoutItems(true);
}

void Basket::addTagToSelectedItems(Tag *tag)
{
	FOR_EACH_ITEM (item)
		item->addTagToSelectedItems(tag);
}

void Basket::removeTagFromSelectedItems(Tag *tag)
{
	FOR_EACH_ITEM (item)
		item->removeTagFromSelectedItems(tag);
}

void Basket::addStateToSelectedItems(State *state)
{
	FOR_EACH_ITEM (item)
		item->addStateToSelectedItems(state);
}

void Basket::changeStateOfSelectedItems(State *state)
{
	FOR_EACH_ITEM (item)
		item->changeStateOfSelectedItems(state);
}

void Basket::removeAllTagsFromSelectedItems()
{
	FOR_EACH_ITEM (item)
		item->removeAllTagsFromSelectedItems();
}

bool Basket::selectedItemsHaveTags()
{
	FOR_EACH_ITEM (item)
		if (item->selectedItemsHaveTags())
			return true;
	return false;
}

QColor Basket::backgroundColor()
{
	if (m_backgroundColorSetting.isValid())
		return m_backgroundColorSetting;
	else
		return KGlobalSettings::baseColor();
}

QColor Basket::textColor()
{
	if (m_textColorSetting.isValid())
		return m_textColorSetting;
	else
		return KGlobalSettings::textColor();
}

void Basket::unbufferizeAll()
{
	Item *item = m_firstItem;
	while (item) {
		item->unbufferizeAll();
		item = item->next();
	}
}

bool Basket::hasTextInEditor()
{
	if (!isDuringEdit() || !redirectEditActions())
		return false;

	if (m_editor->textEdit())
		return ! m_editor->textEdit()->text().isEmpty();
	else if (m_editor->lineEdit())
		return ! m_editor->lineEdit()->text().isEmpty();
	else
		return false;
}

bool Basket::hasSelectedTextInEditor()
{
	if (!isDuringEdit() || !redirectEditActions())
		return false;

	if (m_editor->textEdit()) {
		// The following line does NOT work if one letter is selected and the user press Shift+Left or Shift+Right to unselect than letter:
		// Qt misteriously tell us there is an invisible selection!!
		//return m_editor->textEdit()->hasSelectedText();
		return !m_editor->textEdit()->selectedText().isEmpty();
	} else if (m_editor->lineEdit())
		return m_editor->lineEdit()->hasSelectedText();
	else
		return false;
}

bool Basket::selectedAllTextInEditor()
{
	if (!isDuringEdit() || !redirectEditActions())
		return false;

	if (m_editor->textEdit())
		return m_editor->textEdit()->text().isEmpty() || m_editor->textEdit()->text() == m_editor->textEdit()->selectedText();
	else if (m_editor->lineEdit())
		return m_editor->lineEdit()->text().isEmpty() || m_editor->lineEdit()->text() == m_editor->lineEdit()->selectedText();
	else
		return false;
}

void Basket::selectionChangedInEditor()
{
	Global::mainContainer->countSelectedsChanged();
}

void Basket::placeEditor()
{
	if (!isDuringEdit())
		return;

	QFrame    *editorQFrame = dynamic_cast<QFrame*>(m_editor->widget());
	KTextEdit *textEdit     = m_editor->textEdit();
//	QLineEdit *lineEdit     = m_editor->lineEdit();
	Item      *item         = m_editor->item();

	int frameWidth = (editorQFrame ? editorQFrame->frameWidth() : 0);
	int x          = item->x() + item->contentX() + item->content()->xEditorIndent() - frameWidth;
	int y;
	int maxHeight  = QMAX(visibleHeight(), contentsHeight());
	int height, width;

	if (textEdit) {
		x -= 4;
		// Need to do it 2 times, because it's wrong overwise:
		for (int i = 0; i < 2; i++) {
			// FIXME: CRASH: Select all text, press Del or [<--] and editor->removeSelectedText() is called:
			//        editor->sync() CRASH!!
	//		editor->sync();
			y = item->y();
			height = textEdit->contentsHeight() + 2*frameWidth;
			height = QMAX(height, item->height());
			height = QMIN(height, visibleHeight());
			width  = /*item->x() + item->width()*/item->rightLimit() - x + 2*frameWidth;
			if (y + height > maxHeight)
				y = maxHeight - height;
			textEdit->setFixedSize(width, height);
		}
	} else {
		height = item->height();
		width  = item->rightLimit() - x + 2*frameWidth;
		m_editor->widget()->setFixedSize(width, height);
		x -= 1;
		y = item->y();
	}
	addChild(m_editor->widget(), x, y);
	// ensureWidgetVisible(m_editor->widget()):
	//int visibleX = (contentsX() + visibleWidth()) / 2; // Take middle of view, to not scroll in X
	ensureVisible( item->rightLimit(), y + height, 0,0 ); // Bottom-right corner
	ensureVisible( x,                  y,          0,0 ); // Top-left corner
}

void Basket::closeEditor()
{
	if (!isDuringEdit())
		return;

	if (m_redirectEditActions) {
		disconnect( m_editor->widget(), SIGNAL(selectionChanged()), this, SLOT(selectionChangedInEditor()) );
		if (m_editor->textEdit())
			disconnect( m_editor->textEdit(), SIGNAL(textChanged()),               this, SLOT(selectionChangedInEditor()) );
		else if (m_editor->lineEdit())
			disconnect( m_editor->lineEdit(), SIGNAL(textChanged(const QString&)), this, SLOT(selectionChangedInEditor()) );
	}
	m_editor->widget()->disconnect();
	m_editor->widget()->hide();
	m_editor->validate();

	Item *item = m_editor->item();
	item->setWidth(0); // For relayoutItems() to succeed to take care of the change

	// Delete the item if it is now empty:
	if (m_editor->isEmpty() /*&& m_editor->item()->states().count() <= 0*/) {
		item->setSelected(true);
		item->deleteSelectedItems();
		save();
	}

//	delete m_editor; // Crash!
	m_editor = 0;
	m_redirectEditActions = false;

	unlockHovering();
	searchAgain();

// Does not work:
//	if (Settings::playAnimations())
//		item->setOnTop(true); // So if it grew, do not obscure it temporarily while the items below it are moving

	//relayoutItems(true);
	unselectAll();
	doHoverEffects();
//	save();

	emit resetStatusBarText(); // Remove the "Editing. ... to validate." text.

	//if (kapp->activeWindow() == Global::mainContainer)
		setFocus();
}

Item* Basket::theSelectedItem()
{
	if (countSelecteds() != 1) {
		std::cout << "NO SELECTED ITEM !!!!" << std::endl;
		return 0;
	}

	Item *selectedOne;
	FOR_EACH_ITEM (item) {
		selectedOne = item->theSelectedItem();
		if (selectedOne)
			return selectedOne;
	}

	std::cout << "One selected item, BUT NOT FOUND !!!!" << std::endl;

	return 0;
}

void debugSel(ItemSelection* sel, int n = 0)
{
	for (ItemSelection *node = sel; node; node = node->next) {
		for(int i = 0; i < n; i++)
			std::cout << "-";
		std::cout << (node->firstChild ? "Group" : node->item->content()->toText("")) << std::endl;
		if (node->firstChild)
			debugSel(node->firstChild, n+1);
	}
}

ItemSelection* Basket::selectedItems()
{
	ItemSelection selection;

	FOR_EACH_ITEM (item)
		selection.append(item->selectedItems());

	if (!selection.firstChild)
		return 0;

	for (ItemSelection *node = selection.firstChild; node; node = node->next)
		node->parent = 0;

	// If the top-most groups are columns, export only childs of those groups
	// (because user is not consciencious that columns are groups, and don't care: it's not what she want):
	if (selection.firstChild->item->isColumn()) {
		ItemSelection tmpSelection;
		ItemSelection *nextNode;
		for (ItemSelection *node = selection.firstChild; node; node = node->next) {
			for (ItemSelection *subNode = node->firstChild; subNode; subNode = nextNode) {
				nextNode = subNode->next;
				tmpSelection.append(subNode);
				subNode->parent = 0;
				subNode->next = 0;
			}
		}
//		debugSel(tmpSelection.firstChild);
		return tmpSelection.firstChild;
	} else {
//		debugSel(selection.firstChild);
		return selection.firstChild;
	}
}

void Basket::itemEdit(Item *item, bool justAdded, const QPoint &clickedPoint) // TODO: Remove the first parameter!!!
{
	if (!item)
		item = theSelectedItem(); // TODO: Or pick the focused item!
	if (!item)
		return;

	if (isDuringEdit()) {
		closeEditor(); // Validate the itemeditors in KLineEdit that does not intercept Enter key press (and edit is triggered with Enter too... Can conflict)
		return;
	}

	doHoverEffects(item, Item::Content); // Be sure (in the case Edit was triggered by menu or Enter key...): better feedback!
	m_lockedHovering = true;

	//m_editorWidget = item->content()->launchEdit(this);
	m_editor = ItemEditor::editItemContent(item->content(), this);

	if (m_editor->widget()) {
		m_editor->widget()->reparent(viewport(), QPoint(0,0), true);
		addChild(m_editor->widget(), 0, 0);
		placeEditor(); // FIXME: After?
		m_redirectEditActions = m_editor->lineEdit() || m_editor->textEdit();
		if (m_redirectEditActions) {
			connect( m_editor->widget(), SIGNAL(selectionChanged()), this, SLOT(selectionChangedInEditor()) );
			// In case there is NO text, "Select All" is disabled. But if the user press a key the there is now a text:
			// selection has not changed but "Select All" should be re-enabled:
			if (m_editor->textEdit())
				connect( m_editor->textEdit(), SIGNAL(textChanged()),               this, SLOT(selectionChangedInEditor()) );
			else if (m_editor->lineEdit())
				connect( m_editor->lineEdit(), SIGNAL(textChanged(const QString&)), this, SLOT(selectionChangedInEditor()) );
		}
		m_editor->widget()->show();
		m_editor->widget()->raise();
		m_editor->widget()->setFocus();
		connect( m_editor, SIGNAL(askValidation()), this, SLOT(closeEditor()) );
		if (m_editor->textEdit()) {
			connect( m_editor->textEdit(), SIGNAL(textChanged()), this, SLOT(placeEditor()) );
			if (clickedPoint != QPoint()) {
				QPoint pos(clickedPoint.x() - item->x() - item->contentX() + m_editor->textEdit()->frameWidth() + 4   - m_editor->textEdit()->frameWidth(),
				           clickedPoint.y() - item->y()   - m_editor->textEdit()->frameWidth());
				// Do it right before the kapp->processEvents() to not have the cursor to quickly flicker at end (and sometimes stay at end AND where clicked):
				m_editor->textEdit()->placeCursor(pos);
				//textEdit->viewport()->update();
				kapp->processEvents();
				HtmlEditor *htmlEditor = dynamic_cast<HtmlEditor*>(m_editor);
				if (htmlEditor)
					htmlEditor->cursorPositionChanged(); // Does not work anyway :-( (when clicking on a red bold text, the toolbar still show black normal text)
			}
		}
		kapp->processEvents();     // Show the editor toolbar before ensuring the item is visible
		ensureVisibleItem(item);   //  because toolbar can create a new line and then partially hide the item
		emit resetStatusBarText(); // Display "Editing. ... to validate."
	} else {
		// Delete the item user have canceled the addition:
		if ((justAdded && m_editor->canceled()) || m_editor->isEmpty() /*) && m_editor->item()->states().count() <= 0*/) {
			m_editor->item()->setSelected(true);
			m_editor->item()->deleteSelectedItems();
			//relayoutItems(true);
			//doHoverEffects();
			save();
		}
		m_editor = 0;
		unlockHovering();
		searchAgain();
		unselectAll();
//		save();
	}
}

void Basket::itemDelete()
{
	if (redirectEditActions()) {
		if (m_editor->textEdit())
			m_editor->textEdit()->del();
		else if (m_editor->lineEdit())
			m_editor->lineEdit()->del();
		return;
	}

	if (countSelecteds() <= 0)
		return;
	int really = KMessageBox::questionYesNo( this,
		i18n("<qt>Do you really want to delete this item?</qt>",
		     "<qt>Do you really want to delete those <b>%n</b> items?</qt>",
		     countSelecteds()),
		i18n("Delete Item", "Delete Items", countSelecteds())
#if KDE_IS_VERSION( 3, 2, 90 )   // KDE 3.3.x
		, KStdGuiItem::del(), KStdGuiItem::cancel());
#else
		                    );
#endif
	if (really == KMessageBox::No)
		return;

	itemDeleteWithoutConfirmation();
}

void Basket::itemDeleteWithoutConfirmation(bool deleteFilesToo)
{
	Item *item = firstItem();
	Item *next;
	while (item) {
		next = item->next(); // If we delete 'item' on the next line, item->next() will be 0!
		item->deleteSelectedItems(deleteFilesToo);
		item = next;
	}

	relayoutItems(true);
	doHoverEffects();
	save();
}

void Basket::doCopy(CopyMode copyMode)
{
	QClipboard *cb = KApplication::clipboard();
	QClipboard::Mode mode = (copyMode == CopyToSelection ? QClipboard::Selection : QClipboard::Clipboard);

	ItemSelection *selection = selectedItems();
	int countCopied = countSelecteds();
	if (selection->firstStacked()) {
		QDragObject *d = ItemDrag::dragObject(selection, copyMode == CutToClipboard, /*source=*/0); // d will be deleted by QT
//		/*bool shouldRemove = */d->drag();
//		delete selection;
		cb->setData(d, mode); // ItemMultipleDrag will be deleted by QT
//		if (copyMode == CutToClipboard && !item->useFile()) // If useFile(), ItemDrag::dragObject() will delete it TODO
//			item->slotDelete();

		if (copyMode == CutToClipboard)
			itemDeleteWithoutConfirmation(/*deleteFilesToo=*/false);

		switch (copyMode) {
			default:
			case CopyToClipboard: emit postMessage(i18n("Copied item to clipboard.", "Copied items to clipboard.", countCopied)); break;
			case CutToClipboard:  emit postMessage(i18n("Cutted item to clipboard.", "Cutted items to clipboard.", countCopied)); break;
			case CopyToSelection: emit postMessage(i18n("Copied item to selection.", "Copied items to selection.", countCopied)); break;
		}
	}
}

void Basket::itemCopy()
{
	if (redirectEditActions()) {
		if (m_editor->textEdit())
			m_editor->textEdit()->copy();
		else if (m_editor->lineEdit())
			m_editor->lineEdit()->copy();
	} else
		doCopy(CopyToClipboard);
}

void Basket::itemCut()
{
	if (redirectEditActions()) {
		if (m_editor->textEdit())
			m_editor->textEdit()->cut();
		else if (m_editor->lineEdit())
			m_editor->lineEdit()->cut();
	} else
		doCopy(CutToClipboard);
}

void Basket::itemOpen(Item *item)
{
	/*
	GetSelectedItems
	NoSelectedItem || Count == 0 ? return
	AllTheSameType ?
	Get { url, message(count) }
	*/

	// TODO: Open ALL selected items!
	if (!item)
		item = theSelectedItem();
	if (!item)
		return;

	KURL    url     = item->content()->urlToOpen(/*with=*/false);
	QString message = item->content()->messageWhenOpenning(ItemContent::OpenOne /*ItemContent::OpenSeveral*/);
	if (url.isEmpty()) {
		if (message.isEmpty())
			emit postMessage(i18n("Non openable item.") /*"Non openable items."*/);
		else {
			int result = KMessageBox::warningContinueCancel(this, message, /*caption=*/QString::null, KGuiItem(i18n("&Edit"), "edit"));
			if (result == KMessageBox::Continue)
				itemEdit(item);
		}
	} else {
		emit postMessage(message); // "Openning link target..." / "Launching application..." / "Openning item file..."
		// Finally do the opening job:
		QString customCommand = item->content()->customOpenCommand();
		if (customCommand.isEmpty()) {
			KRun *run = new KRun(url);
			run->setAutoDelete(true);
		} else
			KRun::run(customCommand, url);
	}
}

/** Code from bool KRun::displayOpenWithDialog(const KURL::List& lst, bool tempFiles)
  * It does not allow to set a text, so I ripped it to do that:
  */
bool KRun__displayOpenWithDialog(const KURL::List& lst, bool tempFiles, const QString &text)
{
	if (kapp && !kapp->authorizeKAction("openwith")) {
		KMessageBox::sorry(0L, i18n("You are not authorized to open this file.")); // TODO: Better message, i18n freeze :-(
		return false;
	}
	KOpenWithDlg l(lst, text, QString::null, 0L);
	if (l.exec()) {
		KService::Ptr service = l.service();
		if (!!service)
			return KRun::run(*service, lst, tempFiles);
		//kdDebug(250) << "No service set, running " << l.text() << endl;
		return KRun::run(l.text(), lst); // TODO handle tempFiles
	}
	return false;
}

void Basket::itemOpenWith(Item *item)
{
	if (!item)
		item = theSelectedItem();
	if (!item)
		return;

	KURL    url     = item->content()->urlToOpen(/*with=*/true);
	QString message = item->content()->messageWhenOpenning(ItemContent::OpenOneWith /*ItemContent::OpenSeveralWith*/);
	QString text    = item->content()->messageWhenOpenning(ItemContent::OpenOneWithDialog /*ItemContent::OpenSeveralWithDialog*/);
	if (url.isEmpty())
		emit postMessage(i18n("Non openable item.") /*"Non openable items."*/);
	else if (KRun__displayOpenWithDialog(url, false, text))
		emit postMessage(message); // "Openning link target with..." / "Openning item file with..."
}

void Basket::itemSaveAs()
{
//	if (!item)
//		item = theSelectedItem();
//	if (!item)
//		return;
	Item *item = theSelectedItem();

	KURL url = item->content()->urlToOpen(/*with=*/false);
	if (url.isEmpty())
		return;

	QString fileName = KFileDialog::getSaveFileName(url.fileName(), item->content()->saveAsFilters(), this, i18n("Save to File"));
	// TODO: Ask to overwrite !
	if (fileName.isEmpty())
		return;

	// TODO: Convert format, etc. (use ItemContent::saveAs(fileName))
	KIO::copy(url, KURL(fileName));
}

void Basket::itemMoveOnTop()
{
}

void Basket::itemMoveOnBottom()
{
}

void Basket::itemMoveItemUp()
{
}

void Basket::itemMoveItemDown()
{
}

void Basket::linkLookChanged()
{
	Item *item = m_firstItem;
	while (item) {
		item->linkLookChanged();
		item = item->next();
	}
	relayoutItems(true);
}


void Basket::slotCopyingDone(KIO::Job *, const KURL &, const KURL &to, bool, bool)
{
	Item *item = itemForFullPath(to.path());
	DEBUG_WIN << "Copy finished, load item: " + to.path() + (item ? "" : " --- NO CORRESPONDING ITEM");
	if (item != 0L) {
		item->content()->loadFromFile();
//		if (m_focusedItem == item)   // When inserting a new item we ensure it visble
//			ensureVisibleItem(item); //  But after loading it has certainly grown and if it was
	}                                //  on bottom of the basket it's not visible entirly anymore
}

void Basket::slotCopyingDone2(KIO::Job *job)
{
	if (job->error()) {
		DEBUG_WIN << "Copy finished, ERROR";
		return;
	}
	KIO::FileCopyJob *fileCopyJob = (KIO::FileCopyJob*)job;
	Item *item = itemForFullPath(fileCopyJob->destURL().path());
	DEBUG_WIN << "Copy finished, load item: " + fileCopyJob->destURL().path() + (item ? "" : " --- NO CORRESPONDING ITEM");
	if (item != 0L) {
		item->content()->loadFromFile();
//		if (m_focusedItem == item)   // When inserting a new item we ensure it visble
//			ensureVisibleItem(item); //  But after loading it has certainly grown and if it was
	}                                //  on bottom of the basket it's not visible entirly anymore
}


Item* Basket::itemForFullPath(const QString &path)
{
	Item *item = firstItem();
	Item *found;
	while (item) {
		found = item->itemForFullPath(path);
		if (found)
			return found;
		item = item->next();
	}
	return 0;
}
//
void Basket::deleteFiles()
{
//	m_watcher->stopScan();
	Utilities::deleteRecursively(fullPath());
}

QString Basket::cssFontDefinition(const QFont &font, bool onlyFontFamily)
{
	// The font definition:
	QString definition = QString(font.italic() ? "italic " : "") +
	                     QString(font.bold()   ? "bold "   : "") +
	                     QString::number(QFontInfo(font).pixelSize()) + "px ";

	// Then, try to match the font name with a standard CSS font family:
	QString genericFont = "";
	if (definition.contains("serif", false) || definition.contains("roman", false))
		genericFont = "serif";
	// No "else if" because "sans serif" must be counted as "sans". So, the order between "serif" and "sans" is important
	if (definition.contains("sans", false) || definition.contains("arial", false) || definition.contains("helvetica", false))
		genericFont = "sans-serif";
	if (definition.contains("mono",       false) || definition.contains("courier", false) ||
	    definition.contains("typewriter", false) || definition.contains("console", false) ||
	    definition.contains("terminal",   false) || definition.contains("news",    false))
		genericFont = "monospace";

	// Eventually add the generic font family to the definition:
	QString fontDefinition = "\"" + font.family() + "\"";
	if (!genericFont.isEmpty())
		fontDefinition += ", " + genericFont;

	if (onlyFontFamily)
		return fontDefinition;

	return definition + fontDefinition;
}

/** Save an icon to a directory.
  * If an icon with the same name already exist in the destination,
  * it is assumed the icon is already copied, so no action is took.
  * It is optimized so that you can have an empty folder receiving the icons
  * and call copyIcon() each time you encounter one during export process.
  */
QString Basket::copyIcon(const QString &iconName, int size, const QString &destFolder)
{
	if (iconName.isEmpty())
		return "";

	// Sometimes icon can be "favicons/www.kde.org", we replace the '/' with a '_'
	QString fileName = iconName; // QString::replace() isn't const, so I must copy the string before
	fileName = "ico" + QString::number(size) + "_" + fileName.replace("/", "_") + ".png";
	QString fullPath = destFolder + fileName;
	if (!QFile::exists(fullPath))
		DesktopIcon(iconName, size).save(fullPath, "PNG");
	return fileName;
}

/** Done: Sometimes we can call two times copyFile() with the same srcPath and destFolder
  *       (eg. when exporting basket to HTML with two links to same filename
  *            (but not necesary same path, as in "/home/foo.txt" and "/foo.txt") )
  *       The first copy isn't yet started, so the dest file isn't created and this method
  *       returns the same filename !!!!!!!!!!!!!!!!!!!!
  */

QString Basket::copyFile(const QString &srcPath, const QString &destFolder, bool createIt)
{
	QString fileName = Utilities::fileNameForNewFile(KURL(srcPath).fileName(), destFolder);
	QString fullPath = destFolder + fileName;

	if (createIt) {
		// We create the file to be sure another very near call to copyFile() willn't choose the same name:
		QFile file(KURL(fullPath).path());
		if (file.open(IO_WriteOnly))
			file.close();
		// And then we copy the file AND overwriting the file we juste created:
		new KIO::FileCopyJob(
			KURL(srcPath), KURL(fullPath), 0666, /*move=*/false,
			/*overwrite=*/true, /*resume=*/true, /*showProgress=*/true );
	} else
		/*KIO::CopyJob *copyJob = */KIO::copy(KURL(srcPath), KURL(fullPath)); // Do it as before

	return fileName;
}

QValueList<State*> Basket::usedStates()
{
	QValueList<State*> states;
	FOR_EACH_ITEM (item)
		item->usedStates(states);
	return states;
}

QString Basket::saveGradientBackground(const QColor &color, const QFont &font, const QString &folder)
{
	// Construct file name and return if the file already exists:
	QString fileName = "item_background_" + color.name().lower().mid(1) + ".png";
	QString fullPath = folder + fileName;
	if (QFile::exists(fullPath))
		return fileName;

	// Get the gradient top and bottom colors:
	QColor topBgColor;
	QColor bottomBgColor;
	Item::getGradientColors(color, &topBgColor, &bottomBgColor);

	// Draw and save the gradient image:
	int sampleTextHeight = QFontMetrics(font)
	                       .boundingRect(0, 0, /*width=*/10000, /*height=*/0, Qt::AlignAuto | Qt::AlignTop | Qt::WordBreak, "Test text")
	                       .height();
	QPixmap itemGradient(100, sampleTextHeight + Item::ITEM_MARGIN);
	QPainter painter(&itemGradient);
	drawGradient(&painter, topBgColor, bottomBgColor, 0, 0, itemGradient.width(), itemGradient.height(), /*sunken=*/false, /*horz=*/true, /*flat=*/false);
	painter.end();
	itemGradient.save(fullPath, "PNG");

	// Return the name of the created file:
	return fileName;
}

bool tooDark(const QColor &color); // Defined in item.cpp!

void Basket::exportToHTML()
{
	/*KMessageBox::information(this, "Sorry, this action isn't re-available yet. It need to be adapted to the new BasKet Note Pads capabilities.");*/

	// Get the HTML filename:
	ExporterDialog *options = new ExporterDialog(this);
	if (options->exec() == QDialog::Rejected)
		return;

	// TODO: options->erasePreviousFiles() is not used: on pressing OK, a messagebox should show:
	//       "A folder named foobar_files already exist. This folder will contains the data files and images.
	//        Should it be erased?
	//        [Erase] [Cancel]"
	// TODO: options->formatForImpression() was used to put a line between items.
	//       It's now only used to put URL in front of links. TODO: Use @media print { } for that!

	// Open the file to write:
	QString filePath = options->filePath();
	QFile file(filePath);
	if ( ! file.open(IO_WriteOnly) )
		return;
	QTextStream stream(&file);
	stream.setEncoding(QTextStream::UnicodeUTF8);

	// Create and empty the folder:
	QString filesFolderPath = i18n("HTML exportation files folder", "%1_files").arg(filePath) + "/";                  // eg.: "/home/seb/foo.html_files/"
	QString filesFolderName = i18n("HTML exportation files folder", "%1_files").arg(KURL(filePath).fileName()) + "/"; // eg.: "foo.html_files/"
	Utilities::deleteRecursively(filesFolderPath);
	QDir dir;
	dir.mkdir(filesFolderPath);

	// Create sub-folders:
	QString iconsFolderPath  = filesFolderPath + i18n("HTML exportation icons folder",  "icons")  + "/"; // eg.: "/home/seb/foo.html_files/icons/"
	QString iconsFolderName  = filesFolderName + i18n("HTML exportation icons folder",  "icons")  + "/"; // eg.: "foo.html_files/icons/"
	QString imagesFolderPath = filesFolderPath + i18n("HTML exportation images folder", "images") + "/"; // eg.: "/home/seb/foo.html_files/images/"
	QString imagesFolderName = filesFolderName + i18n("HTML exportation images folder", "images") + "/"; // eg.: "foo.html_files/images/"
	QString dataFolderPath   = filesFolderPath + i18n("HTML exportation data folder",   "data")   + "/"; // eg.: "/home/seb/foo.html_files/data/"
	QString dataFolderName   = filesFolderName + i18n("HTML exportation data folder",   "data")   + "/"; // eg.: "foo.html_files/data/"
	dir.mkdir(iconsFolderPath);
	dir.mkdir(imagesFolderPath);
	dir.mkdir(dataFolderPath);

	// Generate basket icons:
	QString basketIcon16 = iconsFolderName + copyIcon(icon(), 16, iconsFolderPath);
	QString basketIcon32 = iconsFolderName + copyIcon(icon(), 32, iconsFolderPath);
	// Generate the [+] image for groups:
	QPixmap expandGroup(Item::EXPANDER_WIDTH, Item::EXPANDER_HEIGHT);
	expandGroup.fill(backgroundColor());
	QPainter painter(&expandGroup);
	Item::drawExpander(&painter, 0, 0, backgroundColor(), /*expand=*/true, this);
	painter.end();
	expandGroup.save(imagesFolderPath + "expand_group.png", "PNG");
	// Generate the [-] image for groups:
	QPixmap foldGroup(Item::EXPANDER_WIDTH, Item::EXPANDER_HEIGHT);
	foldGroup.fill(backgroundColor());
	painter.begin(&foldGroup);
	Item::drawExpander(&painter, 0, 0, backgroundColor(), /*expand=*/false, this);
	painter.end();
	foldGroup.save(imagesFolderPath + "fold_group.png", "PNG");
	// Compute the colors to draw dragient for items:
	QColor topBgColor;
	QColor bottomBgColor;
	Item::getGradientColors(backgroundColor(), &topBgColor, &bottomBgColor);
	// Compute the gradient image for items:
	QString gradientImageFileName = saveGradientBackground(backgroundColor(), QScrollView::font(), imagesFolderPath);
// 	int sampleTextHeight = QFontMetrics(QScrollView::font())
// 	                       .boundingRect(0, 0, /*width=*/10000, /*height=*/0, Qt::AlignAuto | Qt::AlignTop | Qt::WordBreak, "Test text")
// 	                       .height();
// 	QPixmap itemGradient(100, sampleTextHeight + Item::ITEM_MARGIN);
// 	painter.begin(&itemGradient);
// 	drawGradient(&painter, topBgColor, bottomBgColor, 0, 0, itemGradient.width(), itemGradient.height(), /*sunken=*/false, /*horz=*/true, /*flat=*/false);
// 	painter.end();
// 	itemGradient.save(imagesFolderPath + "item_background_" + backgroundColor().name().lower().mid(1) + ".png", "PNG");

	// Output the header:
	QString borderColor = Utilities::mixColor(backgroundColor(), textColor()).name();
	stream <<
		"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n"
		"         \"http://www.w3.org/TR/html4/strict.dtd\">\n"
		"<html>\n"
		" <head>\n"
		"  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n"
		"  <meta name=\"Generator\" content=\"" << kapp->aboutData()->programName() << " " << VERSION << " http://basket.kde.org/\">\n"
		"  <style type=\"text/css\">\n"
		"   h1 { text-align: center; }\n"
		"   img { border: none; vertical-align: middle; }\n"
		"   .basket { background-color: " << backgroundColor().name() << "; border: solid " << borderColor << " 1px; "
		             "font: " << cssFontDefinition(QScrollView::font()) << "; color: " << textColor().name() << "; padding: 1px; width: 100%; }\n"
		"   table.basket { border-collapse: collapse; }\n"
		"   .basket * { padding: 0; margin: 0; }\n"
		"   .basket table { width: 100%; border-spacing: 0; _border-collapse: collapse; }\n"
		"   .column { vertical-align: top; }\n"
		"   .columnHandle { width: " << Item::RESIZER_WIDTH << "px; background: transparent url('" << imagesFolderName << "column_handle.png') repeat-y; }\n"
		"   .group { margin: 0; padding: 0; border-collapse: collapse; width: 100% }\n"
		"   .groupHandle { margin: 0; width: " << Item::GROUP_WIDTH << "px; text-align: center; }\n"
		"   .item { padding: 1px 2px; background: " << bottomBgColor.name() << " url('" << imagesFolderName << gradientImageFileName << "')"
		          " repeat-x; border-top: solid " << topBgColor.name() <<
		          " 1px; border-bottom: solid " << Utilities::mixColor(topBgColor, bottomBgColor).name() <<
		          " 1px; width: 100%; }\n"
		"   .tags { width: 1px; white-space: nowrap; }\n"
		"   .tags img { padding-right: 2px; }\n"
		<< LinkLook::soundLook->toCSS("sound", textColor())
		<< LinkLook::fileLook->toCSS("file", textColor())
		<< LinkLook::localLinkLook->toCSS("local", textColor())
		<< LinkLook::networkLinkLook->toCSS("network", textColor())
		<< LinkLook::launcherLook->toCSS("launcher", textColor())
		<<
		"   .unknown { margin: 1px 2px; border: 1px solid " << borderColor << "; -moz-border-radius: 4px; }\n";
	QValueList<State*> states = usedStates();
	QString statesCss;
	for (State::List::Iterator it = states.begin(); it != states.end(); ++it)
		statesCss += (*it)->toCSS(imagesFolderPath, imagesFolderName, QScrollView::font());
	stream <<
		statesCss <<
		"   .credits { margin: 3px 0 0 0; font-size: 80%; color: " << borderColor << "; }\n"
		"  </style>\n"
		"  <title>" << Utilities::textToHTMLWithoutP(basketName()) << "</title>\n"
		"  <link rel=\"shortcut icon\" type=\"image/png\" href=\"" << basketIcon16 << "\">\n";
	// Create the column handle image:
	QPixmap columnHandle(Item::RESIZER_WIDTH, 50);
	painter.begin(&columnHandle);
	Item::drawInactiveResizer(&painter, 0, 0, columnHandle.height(), backgroundColor(), /*column=*/true);
	painter.end();
	columnHandle.save(imagesFolderPath + "column_handle.png", "PNG");
	//
	if (/*TODO: hasPNG*/true) {
		// Copy a transparent GIF image in the folder, needed for the JavaScript hack:
		QString gifFileName = "spacer.gif";
		QFile transGIF(imagesFolderPath + gifFileName);
		if (transGIF.open(IO_WriteOnly)) {
			QDataStream streamGIF(&transGIF);
			// This is a 1px*1px transparent GIF image:
			const uchar blankGIF[] = {
				0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x0a, 0x00, 0x0a, 0x00,
				0x80, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x21,
				0xfe, 0x15, 0x43, 0x72, 0x65, 0x61, 0x74, 0x65, 0x64, 0x20,
				0x77, 0x69, 0x74, 0x68, 0x20, 0x54, 0x68, 0x65, 0x20, 0x47,
				0x49, 0x4d, 0x50, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x0a, 0x00,
				0x01, 0x00, 0x2c, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x0a,
				0x00, 0x00, 0x02, 0x08, 0x8c, 0x8f, 0xa9, 0xcb, 0xed, 0x0f,
				0x63, 0x2b, 0x00, 0x3b };
			streamGIF.writeRawBytes((const char*)blankGIF, (unsigned int)74);
			transGIF.close();
			stream <<
				"  <!--[if gte IE 5.5000]>\n"
				"  <script>\n"
				"   window.attachEvent(\"onload\", pngFix);\n"
				"   function pngFix() {\n"
				"    for (i = document.images.length - 1; i >= 0; i--) {\n"
				"     x = document.images[i];\n"
				"     if (x.src.substr(x.src.length - 4) == \".png\") {\n"
				"      x.style.filter = \"progid:DXImageTransform.Microsoft.AlphaImageLoader(src=\'\"+x.src+\"')\"\n"
				"      x.src = \"" << imagesFolderName << gifFileName << "\";\n"
				"     }\n"
				"    }\n"
				"   }\n"
				"  </script>\n"
				"  <![endif]-->\n";
		} // else: do not provide IE compatibility
	}
	stream <<
		" </head>\n"
		" <body>\n"
		"  <h1><img src=\"" << basketIcon32 << "\" width=\"32\" height=\"32\" alt=\"\"> " << Utilities::textToHTMLWithoutP(basketName()) << "</h1>\n";

	// If filtering, only export filtered items, inform to the user:
	// TODO: Filtering tags too!!
	// TODO: Make sure only filtered items are exported!
	if (decoration()->searchData().isSearching)
		stream <<
			"  <p>" << i18n("Items matching &quot;%1&quot; search.").arg(Utilities::textToHTMLWithoutP(decoration()->searchData().string)) << "</p>\n";

	if (isColumnsLayout())
		stream <<
			"  <table class=\"basket\">\n"
			"   <tr>\n";
	else
		stream <<
			"  <div class=\"basket\" style=\"position: relative; height: " << contentsHeight() << "px; width: " << contentsWidth() << "px; min-width: 100%;\">\n";

	// output the body:
	HtmlExportData exportData;
	exportData.iconsFolderPath     = iconsFolderPath;
	exportData.iconsFolderName     = iconsFolderName;
	exportData.imagesFolderPath    = imagesFolderPath;
	exportData.imagesFolderName    = imagesFolderName;
	exportData.dataFolderPath      = dataFolderPath;
	exportData.dataFolderName      = dataFolderName;
	exportData.formatForImpression = options->formatForImpression();
	exportData.embedLinkedFiles    = options->embedLinkedFiles();
	exportData.embedLinkedFolders  = options->embedLinkedFolders();

	FOR_EACH_ITEM (item)
		item->exportToHTML(stream, /*indent=*/(isFreeLayout() ? 3 : 4), exportData);

	// Output the footer:
	if (isColumnsLayout())
		stream <<
			"   </tr>\n"
			"  </table>\n";
	else
		stream <<
			"  </div>\n";
	stream << QString(
		"  <p class=\"credits\">%1</p>\n"
		" </body>\n"
		"</html>\n").arg(
			i18n("Made with %1, a KDE tool to take notes and keep a full range of data on hand.")
				.arg("<a href=\"http://basket.kde.org/\">%1</a> %2")
				.arg(kapp->aboutData()->programName(), VERSION));
	file.close();
}





#if 0

#include <qlayout.h>
#include <qvbox.h>
#include <qstring.h>
#include <qpixmap.h>
#include <qcolor.h>
#include <kpopupmenu.h>
#include <kurllabel.h>
#include <qcheckbox.h>
#include <qpalette.h>
#include <qcursor.h>
#include <qaction.h>
#include <kstdaccel.h>
#include <kglobalsettings.h>
#include <qevent.h>

#include <kapplication.h>
#include <kaboutdata.h>
#include <qinputdialog.h>
#include <qdragobject.h>
#include <kurldrag.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kmimetype.h>
#include <kfiledialog.h>
#include <qdir.h>
#include <kiconloader.h>
#include <qregexp.h>
#include <qfileinfo.h>

#include <qstringlist.h>
#include <qdir.h>
#include <kurl.h>
#include <krun.h>
#include <kmessagebox.h>
#include <kdeversion.h>

#include "kdirwatch.h"
#include <qstringlist.h>
#include <klineedit.h>

#include <config.h>
#include <qtextcodec.h>

#include "basket.h"
#include "item.h"
#include "itemfactory.h"
#include "variouswidgets.h"
#include "linklabel.h"
#include "global.h"
#include "container.h"
#include "xmlwork.h"
#include "settings.h"
#include "popupmenu.h"
#include "debugwindow.h"
#include "exporterdialog.h"


/** Basket */

const int Basket::c_updateTime = 200;

Basket::Basket(QWidget *parent, const QString &folderName, const char *name, WFlags fl)
 : QScrollView(parent, name, fl)
{
//...
	// Create m_watcher before load() because mirrored files should be added to it
	m_watcher = new KDirWatch(this);
	m_watcher->addDir(fullPath(), true); // Watch all files modifications
	connect( m_watcher,     SIGNAL(dirty(const QString&)),   this, SLOT(slotModifiedFile(const QString&)) );
	connect( m_watcher,     SIGNAL(created(const QString&)), this, SLOT(slotCreatedFile(const QString&))  );
	connect( m_watcher,     SIGNAL(deleted(const QString&)), this, SLOT(slotDeletedFile(const QString&))  );
	connect(&m_updateTimer, SIGNAL(timeout()),               this, SLOT(slotUpdateItems())                );

	m_watcher->stopScan();
	load(); // We disable Dir scan during load, in case files are created (when importing launchers)
	m_watcher->startScan();
	resetSearch();
}





void Basket::setLocked(bool lock)
{
	if (lock == m_isLocked)
		return;

	if (isDuringEdit())
		closeEditor();

	m_isLocked = lock;
	for (Item *it = firstItem(); it != 0L; it = it->next())
		it->lockedChanged(lock);

//	m_emptyHelp->setEnabled(!lock);
}

// Remove the item from the basket and delete the associated file
// If the item mirror a file, it will ask before deleting or not the file
// But if askForMirroredFile is false, it willn't ask NOR delete the MIRRORED file
//  (it will anyway delete the file if it is not a mirror)
void Basket::delItem(Item *item, bool askForMirroredFile)
{
//...
	if (hasFocus())
		focusAnItem();       // We need item->next() and item->previous() here [BUT deleted item should be hidden]
	if (item->isSelected())
		item->setSelected(false); //removeSelectedItem();

	relayoutItems();
	recolorizeItems();
	resetInsertTo();         // If we delete the first or the last, pointer to it is invalid
	save();

	if (item == m_startOfShiftSelectionItem)
		m_startOfShiftSelectionItem = 0L;

	if (isDuringEdit() && m_editor->editedItem() == item)
		closeEditor(false);
//...
}


void Basket::contextMenuEvent(QContextMenuEvent *event)
{
	QPopupMenu *menu = Global::mainContainer->popupMenu("item_popup");

	if (event->reason() == QContextMenuEvent::Mouse) {
		Item *itemUnderMouse = itemAtPosition( QPoint(event->pos().x() + contentsX(), event->pos().y() + contentsY()) );
		if (itemUnderMouse == 0L)
			unselectAll();
		else if ( ! itemUnderMouse->isSelected() )
			unselectAllBut(itemUnderMouse);
		menu->exec(event->globalPos());
	} else {
		if (countShown() == 0) {
			QRect basketRect( mapToGlobal(QPoint(0,0)), size() );
			PopupMenu::execAtRectCenter(*menu, basketRect); // Popup at center or the basket
		} else {
			if ( ! m_focusedItem->isSelected() )
				unselectAllBut(m_focusedItem);
			// Popup at bottom (or top) of the focused item, if visible :
			PopupMenu::execAtRectBottom(*menu, itemRect(m_focusedItem), true);
		}
	}
}

QRect Basket::itemRect(Item *item) // TODO: QRect Basket::itemVisibleRect(Item *item)
{
//	QRect rect( contentsToViewport(QPoint(0,item->y())),    // Doesn't work
//	            QSize(item->width(),item->height())      );
	QRect rect( 0 - contentsX(), item->y() - contentsY(), item->width(), item->height() );
	QPoint basketPoint = mapToGlobal(QPoint(0,0));
	rect.moveTopLeft( rect.topLeft() + basketPoint + QPoint(frameWidth(), frameWidth()) );

	// Now, rect contain the global item rectangle on the screen.
	// We have to clip it by the basket widget :
	if (rect.bottom() > basketPoint.y() + visibleHeight() + 1) { // Bottom too... bottom
		rect.setBottom( basketPoint.y() + visibleHeight() + 1);
		if (rect.height() <= 0) // Have at least one visible pixel of height
			rect.setTop(rect.bottom());
	}
	if (rect.top() < basketPoint.y() + frameWidth()) { // Top too... top
		rect.setTop( basketPoint.y() + frameWidth());
		if (rect.height() <= 0)
			rect.setBottom(rect.top());
	}
	if (rect.right() > basketPoint.x() + visibleWidth() + 1) { // Right too... right
		rect.setRight( basketPoint.x() + visibleWidth() + 1);
		if (rect.width() <= 0) // Have at least one visible pixel of width
			rect.setLeft(rect.right());
	}
	if (rect.left() < basketPoint.x() + frameWidth()) { // Left too... left
		rect.setLeft( basketPoint.x() + frameWidth());
		if (rect.width() <= 0)
			rect.setRight(rect.left());
	}

	return rect;
}

// Calculate where to paste or drop
void Basket::computeInsertPlace(const QPoint &cursorPosition)
{
	int y = cursorPosition.y();

	if (countShown() == 0)
		return;

	// TODO: Memorize the last hovered item to avoid a new computation on dragMoveEvent !!
	//       If the mouse is not over the last item, compute which new is :
	// TODO: Optimization : start from m_insertAtItem and compare y position to search before or after (or the same)
	for (Item *it = firstItem(); it != 0L; it = it->next())
		if ( (it->isShown()) && (it->y() + it->height() >= y) && (it->y() < y) ) {
			int center = it->y() + (it->height() / 2);
			m_insertAtItem = it;
			m_insertAfter  = y > center;
			return;
		}
	// Else, there is at least one shown item but cursor hover NO item, so we are after the last shown item
	m_insertAtItem = lastShownItem();
	m_insertAfter  = true;

	// Code for rectangular items :
	/*QRect globalRect = it->rect();
	globalRect.moveTopLeft(it->pos() + contentsY());
	if ( globalRect.contains(curPos) ) {
		it->doInterestingThing();
	}*/
}

void Basket::dragMoveEvent(QDragMoveEvent* event)
{
//	m_isDuringDrag = true;

	if (isLocked())
		return;

//	FIXME: viewportToContents does NOT work !!!
//	QPoint pos = viewportToContents(event->pos());
	QPoint pos( event->pos().x() + contentsX(), event->pos().y() + contentsY() );

//	if (insertAtCursorPos())
		computeInsertPlace(pos);

	showFrameInsertTo();
	acceptDropEvent(event);

	// A workarround since QScrollView::dragAutoScroll seem to have no effect :
	ensureVisible(event->pos().x() + contentsX(), event->pos().y() + contentsY(), 30, 30);
//	QScrollView::dragMoveEvent(event);
}

void Basket::dropEvent(QDropEvent *event)
{
	m_isDuringDrag = false;
	emit resetStatusBarText();

	if (isLocked())
		return;

	ItemFactory::dropItem( event, this, true, event->action(), dynamic_cast<Item*>(event->source()) );
	// TODO: need to know if we really inserted an (or several!!!!) item !!!
	ensureVisibleItem(lastInsertedItem());
	unselectAllBut(lastInsertedItem());
	setFocusedItem(lastInsertedItem());

	resetInsertTo();
}


void Basket::contentsMousePressEvent(QMouseEvent *event)
{
	if (event->button() & Qt::LeftButton) { // Clicked an empty area of the basket
		if ( ! isDuringEdit() ) // Do nothing when clicked in edit mode
			unselectAll();
		setFocus();
	}
	// if MidButton BUT NOT ALT PRESSED to allow Alt+middleClick to launch actions
	//              Because KWin handle Alt+click : it's allow an alternative to alt+click !
	if ( (event->button() & Qt::MidButton) && (event->state() == 0) ) {
//		if (insertAtCursorPos())
			computeInsertPlace(event->pos());
		pasteItem(QClipboard::Selection);
		event->accept();
	} else if ( (Settings::middleAction() != 0) &&
	            (event->button() & Qt::MidButton) && (event->state() == Qt::ShiftButton) ) {
//		if (insertAtCursorPos())
			computeInsertPlace(event->pos());
		// O:Nothing ; 1:Paste ; 2:Text ; 3:Html ; 4:Image ; 5:Link ; 6:Launcher ; 7:Color
		// TODO: 8:Ask (with a popup menu)
		Item::Type type = (Item::Type)0;
		switch (Settings::middleAction()) {
			case 1: pasteItem();           break;
			case 2: type = Item::Text;     break;
			case 3: type = Item::Html;     break;
			case 4: type = Item::Image;    break;
			case 5: type = Item::Link;     break;
			case 6: type = Item::Launcher; break;
			case 7: type = Item::Color;    break;
		}
		if (type != 0)
			Global::mainContainer->insertEmpty(type);
		event->accept();
	}
}



void Basket::focusInEvent(QFocusEvent*)
{
	if (m_editor != 0L) // It arrive toolbar of another editor stay shown... FIXME: Should be solved in 0.5.1
		closeEditor();

	focusAnItem();      // hasFocus() is true at this stage, item will be focused
}

void Basket::focusOutEvent(QFocusEvent*)
{
	if (m_focusedItem != 0L)
		m_focusedItem->setFocused(false);
}

void Basket::processActionAsYouType(QKeyEvent *event)
{
	if (Global::debugWindow)
		*Global::debugWindow << QString("Key press (%1)").arg(event->text());

	if ( (kapp->focusWidget() != this) ||    // Do not receive child keyPress events
	     (event->key() == Qt::Key_Escape) || // This key generate a text() !
	     (event->text().isEmpty()) ||        // It should be a text (not a modifier...)
	     (event->state() & AltButton) ||     // If user pressed an unexisting shortcut
	     (event->state() & MetaButton) ||    //  or accelerator key, do nothing
	     ((event->state() & ControlButton) && (event->key() != Qt::Key_Backspace)) ) // But allow Ctrl+BreakSpace
		return;                                                                        //  for search bar

	DecoratedBasket *decoration = (DecoratedBasket*)parent();
	SearchBar       *searchBar  = decoration->searchBar();

	int  action         = Settings::writingAction();
	bool takeCareOfText = true;
	if ( event->text().startsWith(",") &&
	     (Settings::writingCommaAction() != 0) && // Do the same, so don't process the ',' as a special case!
		 (Settings::writingCommaAction() != Settings::writingAction()) ) { // Do the same too!
		action = Settings::writingCommaAction();
		takeCareOfText = false;
	}
	if (decoration->isSearchBarShown()) {
		action = 1;
		takeCareOfText = true;
	}

	// O:Nothing ; 1:Search ; 2:Text ; 3:Html ; 4:Link
	if (action == 1) {
		if ((event->key() == Qt::Key_Backspace) && !decoration->isSearchBarShown())
			return; // Do not show and then hide bar (avoid two basket savings)
		if ( ! decoration->isSearchBarShown() )
			Global::mainContainer->showHideSearchBar(true, false);
		if (takeCareOfText)
			kapp->sendEvent(searchBar->lineEdit(), event);
		if (searchBar->lineEdit()->text().isEmpty() && takeCareOfText) // takeCareOfText: comma just show the (empty) bar
			Global::mainContainer->showHideSearchBar(false);
	} else if (action > 1) {
		if ( (isLocked()) ||                        // Do not insert item if locked
		     (event->key() == Qt::Key_Backspace) || // Or if key is backspace !
		     (m_stackedKeyEvent != 0L) )            // Sometimes, insert link dialog can take time to appear
			return;                                 //  and lot of links would be created otherwise.
		Item::Type type = Item::Text;
		if (action == 3)
			type = Item::Html;
		if (action == 4)
			type = Item::Link;
		if (takeCareOfText)
			m_stackedKeyEvent = new QKeyEvent(*event);
		else
			m_stackedKeyEvent = 0L;
		Global::mainContainer->insertEmpty(type);
	}
}

void Basket::keyPressEvent(QKeyEvent *event)
{
	if (countShown() == 0) {
		processActionAsYouType(event);
		//event->ignore(); // Important !!
		return;
	}

	Item *toFocus = 0L;

	switch (event->key()) {
		case Qt::Key_Down:
			toFocus = nextShownItemFrom(m_focusedItem, 1);
			break;
		case Qt::Key_Up:
			toFocus = nextShownItemFrom(m_focusedItem, -1);
			break;
		case Qt::Key_PageDown:
			toFocus = nextShownItemFrom(m_focusedItem, 10);
			if (toFocus == 0L)
				toFocus = lastShownItem();
			break;
		case Qt::Key_PageUp:
			toFocus = nextShownItemFrom(m_focusedItem, -10);
			if (toFocus == 0L)
				toFocus = firstShownItem();
			break;
		case Qt::Key_Home:
			toFocus = firstShownItem();
			break;
		case Qt::Key_End:
			toFocus = lastShownItem();
			break;
		case Qt::Key_Left:   // Those cases do not move focus to another item...
			scrollBy(-30, 0);
			return;
		case Qt::Key_Right:
			scrollBy( 30, 0);
			return;
		case Qt::Key_Space:  // This case do not move focus to another item...
			if (m_focusedItem != 0L) {
				m_focusedItem->setSelected( ! m_focusedItem->isSelected() );
				recolorizeItems();
				event->accept();
			} else
				event->ignore();
			return;          // ... so we return after the process
		default:
			processActionAsYouType(event);
			return;
	}

	if (toFocus == 0L) { // If no direction keys have been pressed OR reached out the begin or end
		event->ignore(); // Important !!
		return;
	}

	if (event->state() & Qt::ShiftButton) { // Shift+arrowKeys selection
		if (m_startOfShiftSelectionItem == 0L)
			m_startOfShiftSelectionItem = toFocus;
		selectRange(m_startOfShiftSelectionItem, toFocus);
		setFocusedItem(toFocus);
		ensureVisibleItem(toFocus);
		event->accept();
		return;
	} else /*if (toFocus != m_focusedItem)*/ {  // Move focus to ANOTHER item...
		m_startOfShiftSelectionItem = toFocus;
		setFocusedItem(toFocus);
		ensureVisibleItem(toFocus);
		if ( ! (event->state() & Qt::ControlButton) )       // ... select only current item if Control
			unselectAllBut(m_focusedItem);
		event->accept();
		return;
	}

	event->ignore(); // Important !!
}

void Basket::ensureVisibleItem(Item *item)
{
	if ( ! item->isShown() ) // It's logical.
		return;

	int visibleX = (contentsX() + visibleWidth()) / 2; // Take middle of view, to not scroll in X
	ensureVisible( visibleX, item->y() + item->height(), 0,0 );
	ensureVisible( visibleX, item->y(),                  0,0 );
}

/** Select a range of items and deselect the others.
  * The order between start and end has no importance (end could be before start)
  */
void Basket::selectRange(Item *start, Item *end)
{
	Item *cur;
	Item *realEnd = 0L;

	// Avoid crash when start (or end) is null
	if (start == 0L)
		start = end;
	else if (end == 0L)
		end = start;
	// And if *both* are null
	if (start == 0L) {
		unselectAll();
		return;
	}
	// In case there is only one item to select
	if (start == end) {
		unselectAllBut(start);
		return;
	}

	// Search the REAL first (and deselect the others before it) :
	for (cur = firstItem(); cur != 0L; cur = cur->next()) {
		if (cur == start || cur == end)
			break;
		cur->setSelected(false);
	}

	// Select the items after REAL start, until REAL end :
	if (cur == start)
		realEnd = end;
	else if (cur == end)
		realEnd = start;

	for (/*cur = cur*/; cur != 0L; cur = cur->next()) {
		cur->setSelected(cur->isShown()); // Select all items in the range, but only if they are shown
		if (cur == realEnd)
			break;
	}

	// Deselect the leaving items :
	for (cur = cur->next(); cur != 0L; cur = cur->next())
		cur->setSelected(false);

	recolorizeItems();
}

Item* Basket::nextShownItemFrom(Item *item, int step)
{
	if (step > 0) {
		while (step && item) {
			item = item->next();
			if (item && item->isShown())
				step--;
		}
	} else {
		while (step && item) {
			item = item->previous();
			if (item && item->isShown())
				step++;
		}
	}
	return item;
}

void Basket::clicked(Item *item, bool controlPressed, bool shiftPressed)
{
	if (shiftPressed) {
		if (m_startOfShiftSelectionItem == 0L)
			m_startOfShiftSelectionItem = item;
		selectRange(m_startOfShiftSelectionItem, item);
	} else {
		m_startOfShiftSelectionItem = item;
		if (controlPressed) {
			item->setSelected( ! item->isSelected() );
			recolorizeItems();
		} else
			unselectAllBut(item);
	}

	setFocusedItem(item);
}


void Basket::exportToHTML()
{
	// Get the HTML filename:
	ExporterDialog *options = new ExporterDialog(this);
	if (options->exec() == QDialog::Rejected)
		return;

	// Open the file to write:
	QString filePath = options->filePath();
	QFile file(filePath);
	if ( ! file.open(IO_WriteOnly) )
		return;

	// Smarter export: export only things that should be exported
	// (don't include unneeded clases, or javascript hacks if there is no tranparent PNG...):
	bool hasFolder       = false; // If we should create a folder to put all files!            // TODO: Create the folder anyway and if empty at end, delete it
	bool hasPNG          = false;                                                              // TODO: At end, if no PNG exist... ther isn't one :)
	// LinkLooks:
	bool hasSound        = false;
	bool hasFiles        = false;
	bool hasLocalLinks   = false;
	bool hasNetworkLinks = false;
	bool hasLaunchers    = false;
	bool hasColors       = false; // If we have at least one Color item, we can put the .color class in the HTML

	if (!icon().isEmpty()) {
		hasFolder = true;
		hasPNG    = true;
	}
	LinkLook *look;
	for (Item *it = firstShownItem(); it != 0L; it = it->next())
		if (it->isShown()) {
			// Check:
			switch (it->type()) {
				case Item::Text:
				case Item::Html:
					break;
				case Item::Image:
					hasFolder = true;
					if (it->fileName().endsWith(".png")) // TODO: If FORMAT is "PNG"
						hasPNG = true;
					break;
				case Item::Animation:
					hasFolder = true;
					break;
				case Item::Sound:
					hasFolder = true;
					if (LinkLook::soundLook->showIcon())
						hasPNG = true;
					hasSound = true;
					break;
				case Item::File:
					hasFolder = true;
					if (LinkLook::fileLook->showIcon() && true) // We're assuming every files have an icon
						hasPNG = true;
					hasFiles = true;
					break;
				case Item::Link:
					look = LinkLook::lookForURL(it->url());
					if (look->showIcon() && !it->icon().isEmpty()) {
						hasPNG = true;
						hasFolder = true;
					}
					if (look == LinkLook::localLinkLook) {
						hasLocalLinks = true;
						if (!hasFiles) { // optimisation: don't verify if already true:
							QFileInfo fInfo(it->url().url());
							if ( (options->embedLinkedFiles()   && fInfo.isFile()) ||
							     (options->embedLinkedFolders() && fInfo.isDir())    )
								hasFiles = true;
						}
					} else
						hasNetworkLinks = true;
					break;
				case Item::Launcher:
					hasFolder = true;
					if (LinkLook::launcherLook->showIcon() && !it->service()->icon().isEmpty())
						hasPNG = true;
					hasLaunchers = true;
					break;
				case Item::Color:
					hasColors = true;
					break;
				case Item::Unknown:
					break;
			}
			if (it == lastShownItem())
				break;
		}

	// Create the folder only if files should be created:
	QString folderPath = i18n("HTML exportation folder", "%1_files").arg(filePath) + "/";
	QString folderName = i18n("HTML exportation folder", "%1_files").arg(KURL(filePath).fileName()) + "/";
	// eg.: folderPath == "/home/seb/exportFileName.html_files/"
	//      folderName == "exportFileName.html_files/"

	QDir dir("");
	if (hasFolder) {
		dir.mkdir(folderPath);
		if (options->erasePreviousFiles())
			Utilities::deleteRecursively(folderPath);
	}

	QString basketIcon16 = copyIcon(icon(), 16, folderPath);
	QString basketIcon32 = copyIcon(icon(), 32, folderPath);
	if ( ! icon().isEmpty() ) {
		basketIcon16 = folderName + basketIcon16;
		basketIcon32 = folderName + basketIcon32;
	}

	QColor foreground = KApplication::palette().active().foreground();
	QTextStream stream(&file);
	stream.setEncoding(QTextStream::UnicodeUTF8);
	stream << QString(
		"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n"
		"         \"http://www.w3.org/TR/html4/strict.dtd\">\n"
		"<html>\n"
		"  <head>\n"
		"    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n"
		"    <meta name=\"Generator\" content=\"%1 %2 http://basket.kde.org/\">\n"
		"    <style type=\"text/css\">\n"
		"      h1 { text-align: center; }\n"
		"      img { border: none; vertical-align: middle; }\n")
			// transform eg. "charset=ISO 8859-15" to content="charset=ISO-8859-15"
			//.arg(/*QString(QTextCodec::codecForLocale()->name()).replace(" ", "-"), */VERSION);
			.arg(kapp->aboutData()->programName(), VERSION);
	QString colorName = color().name();
	QString textAlign = (hAlign() == 0 ? "left" : (hAlign() == 1 ? "center" : "right" ));
	// Set the default font (for all the basket), and try to get a CSS generic family for it (serif, sans-serif...):
	QFont appFont = kapp->font();
	QString font = QString((appFont.italic() ? "italic " : "")) +
	               QString((appFont.bold()   ? "bold "   : "")) +
	               QString::number(QFontInfo(appFont).pixelSize()) + "px " + appFont.family();
	QString genericFont = "";
	if (font.contains("serif", false) || font.contains("roman", false))
		genericFont = "serif";
	if (font.contains("sans", false) || // No "else if" because "sans serif" must
	    font.contains("arial", false) || font.contains("helvetica", false))
		genericFont = "sans-serif"; // be counted as "sans". So, the order between "serif" and "sans" is important
	if (font.contains("mono", false) || font.contains("courier", false) ||
	    font.contains("typewriter", false) || font.contains("console", false) ||
	    font.contains("terminal", false) || font.contains("news", false))
		genericFont = "monospace";
	if (!genericFont.isEmpty())
		font += QString(", ") + genericFont;

	QString itemExtendedStyle;
	if (options->formatForImpression())
		itemExtendedStyle = " border-bottom: solid black 1px;";

/*	if (showCheckBoxes())
		stream <<
			"      .basket { border-collapse: collapse; width: 100%; border: solid black 1px;\n"
			"                font: " << font << "; color: " << foreground.name() << "; }\n"
			"      .item { background-color: " << colorName <<
			            "; margin: 0; text-align: " << textAlign << ";" << itemExtendedStyle <<" }\n"
			"      td, input { padding: 0; margin: 0; }\n"
			"      input { margin-" << (hAlign() != 2 ? "left" : "right") << ": 2px; }\n"
			"      .content { width: 100%; padding: 2px; }\n";
	else*/
		stream <<
			"      .basket { border: solid black 1px; font: " << font << "; color: " << foreground.name() << "; }\n"
			"      .item { background-color: " << colorName <<
			            "; margin: 0; padding: 2px; text-align: " << textAlign << ";" << itemExtendedStyle <<" }\n";

	bool needAltClass = false;
/*	if (color() != altColor() && m_countShown > 1 && !options->formatForImpression()) { // Don't use alternate color for print
		stream << "      .alt { background-color: " << altColor().name() << "; }\n";
		needAltClass = true;
	}
*/
	if (hasColors)       stream << "      .color { font-weight: bold; }\n";
	if (hasSound)        stream << LinkLook::soundLook->toCSS("sound");
	if (hasFiles)        stream << LinkLook::fileLook->toCSS("file");
	if (hasLocalLinks)   stream << LinkLook::localLinkLook->toCSS("local");
	if (hasNetworkLinks) stream << LinkLook::networkLinkLook->toCSS("network");
	if (hasLaunchers)    stream << LinkLook::launcherLook->toCSS("launcher");
	stream <<
		"      .credits { margin: 0; font-size: 80%; }\n"
		"    </style>\n"
		"    <title>" << textToHTMLWithoutP(name()) << "</title>\n";
	if ( ! icon().isEmpty() )
		stream << "    <link rel=\"shortcut icon\" type=\"image/png\" href=\"" << basketIcon16 << "\">\n";
	// Put JavaScript hack to get transparent PNG working in Internet Explorer:
	if (hasPNG) {
		// Copy a transparent GIF image in the folder, needed for the JavaScript hack:
		QString gifFileName = Utilities::fileNameForNewFile("spacer.gif", folderPath);
		QFile transGIF(folderPath + gifFileName);
		if (transGIF.open(IO_WriteOnly)) {
			QDataStream streamGIF(&transGIF);
			// This is a 1px*1px transparent GIF image:
			const uchar blankGIF[] = {
				0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x0a, 0x00, 0x0a, 0x00,
				0x80, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x21,
				0xfe, 0x15, 0x43, 0x72, 0x65, 0x61, 0x74, 0x65, 0x64, 0x20,
				0x77, 0x69, 0x74, 0x68, 0x20, 0x54, 0x68, 0x65, 0x20, 0x47,
				0x49, 0x4d, 0x50, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x0a, 0x00,
				0x01, 0x00, 0x2c, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x0a,
				0x00, 0x00, 0x02, 0x08, 0x8c, 0x8f, 0xa9, 0xcb, 0xed, 0x0f,
				0x63, 0x2b, 0x00, 0x3b };
			streamGIF.writeRawBytes((const char*)blankGIF, (unsigned int)74);
			transGIF.close();

			stream <<
				"    <!--[if gte IE 5.5000]>\n"
				"    <script>\n"
				"      window.attachEvent(\"onload\", pngFix);\n"
				"      function pngFix() {\n"
				"        for (i = document.images.length - 1; i >= 0; i--) {\n"
				"          x = document.images[i];\n"
//				"          //if (x.className == 'pngAlpha') {\n"
				"          if (x.src.substr(x.src.length - 4) == \".png\") {\n"
				"            x.style.filter = \"progid:DXImageTransform.Microsoft.AlphaImageLoader(src=\'\"+x.src+\"')\"\n"
				"            x.src = \"" << folderName << gifFileName << "\";\n"
				"          }\n"
				"        }\n"
				"      }\n"
				"    </script>\n"
				"    <![endif]-->\n";
		} // else: do not provide IE compatibility
	}
	stream << QString(
		"  </head>\n"
		"  <body>\n");
	if ( ! icon().isEmpty() )
		stream << QString(
			"    <h1><img src=\"%1\" width=\"32\" height=\"32\" alt=\"\"> %2</h1>\n")
				.arg(basketIcon32, textToHTMLWithoutP(name()));
	else
		stream << QString("    <h1>%1</h1>\n").arg(textToHTMLWithoutP(name()));

	if ( ((DecoratedBasket*)parent())->searchData().isSearching )
		stream << QString("    <p>%1</p>\n").arg("Items matching &quot;%1&quot; search.")
			.arg(textToHTMLWithoutP( ((DecoratedBasket*)parent())->searchData().string ));

//	if (showCheckBoxes())
//		stream << "    <table class=\"basket\">\n";
//	else
		stream << "    <div class=\"basket\">\n";

	int i = 0;
	LinkLook *linkLook;
	QString   linkURL;
	QString   linkTitle;
	QString   linkIcon;
	// Browse all SHOWN items :
	for (Item *it = firstShownItem(); it != 0L; it = it->next())
		if (it->isShown()) {
			i++;
			// Compute item's class ("item color", "item alt color", "item file"...):
			QString cssClass = "item";
			if (needAltClass && !(i % 2))
				cssClass += " alt";
			if (it->type() == Item::Color)
				cssClass += " color";
			else if (it->useLinkLabel()) {
				if (it->type() == Item::Sound) {
					linkLook = LinkLook::soundLook;
					cssClass += " sound";
				} else if (it->type() == Item::File) {
					linkLook = LinkLook::fileLook;
					cssClass += " file";
				} else if (it->type() == Item::Link) {
					linkLook = LinkLook::lookForURL(it->url());
					cssClass += (linkLook == LinkLook::localLinkLook ? " local" : " network");
				} else if (it->type() == Item::Launcher) {
					linkLook = LinkLook::launcherLook;
					cssClass += " launcher";
				}
			}
			// Compute text item additionnal style:
			//QString textFontType = "";
			/*if (it->type() == Item::Text) {
				switch (it->textFontType()) { // Set font if non-default
					case 1:  textFontType = "font-family: sans-serif;"; break;
					case 2:  textFontType = "font-family: serif;";      break;
					case 3:  textFontType = "font-family: monospace;";  break;
				}
				if (it->textColor() != foreground) { // Set/Add color if non-default
					if (!textFontType.isEmpty())
						textFontType += " ";
					textFontType += QString("color: %1;").arg(it->textColor().name());
				}
				if (!textFontType.isEmpty()) // If at least one thing non-default, set the style
					textFontType = QString(" style=\"%1\"").arg(textFontType);
			}*/
			// Compute annotations:
			QString attrAnnotations = "";
			if ( ! it->annotations().isEmpty() )
				attrAnnotations = QString(" title=\"%1\"")
					.arg(it->annotations().replace("\"", "''"));
			// Output checkbox (if items are aligned at left or center):
/*			if (showCheckBoxes()) {
				stream << QString("      <tr class=\"%1\"%2>\n").arg(cssClass, attrAnnotations);
				if (hAlign() != 2)
					stream << QString(
						"        <td><input type=\"checkbox\"%1 disabled></td>\n")
							.arg(it->isChecked() ? " checked" : "");
			// Output open tag of content:
				stream << QString("        <td class=\"content\">");//%1>").arg(textFontType);
			} else*/
				stream << QString("      <p class=\"%1\""/*%2*/"%3>").arg(cssClass, /*textFontType, */attrAnnotations);
			// Output item content OR compute values for file oriented items:
			switch (it->type()) {
				case Item::Text:
					stream << textToHTMLWithoutP(it->text());
					break;
				case Item::Html:
					stream << htmlToParagraph(it->html());
					break;
				case Item::Image:
				case Item::Animation:
					{
						QSize size = (it->type() == Item::Image ? it->pixmap()->size() : it->movie()->framePixmap().size() );
						stream << QString("<img src=\"%1\" width=\"%2\" height=\"%3\" alt=\"\">")
							.arg( folderName + copyFile(it->fullPath(), folderPath, true),
							      QString::number(size.width()), QString::number(size.height()) );
					}
					break;
				case Item::Sound:
				case Item::File:
					linkURL   = folderName + copyFile(it->fullPath(), folderPath, true);
					linkTitle = it->fileName();
					linkIcon  = ItemFactory::iconForURL(KURL(it->fullPath()));
					break;
				case Item::Launcher:
					linkURL   = folderName + copyFile(it->fullPath(), folderPath, true);
					linkTitle = it->service()->name();
					linkIcon  = it->service()->icon();
					break;
				case Item::Link:
					linkTitle = it->title();
					linkIcon  = it->icon();
					{
						QFileInfo fInfo(it->url().path());
						DEBUG_WIN << it->url().path()
						          << "IsFile:" + QString::number(fInfo.isFile())
						          << "IsDir:"  + QString::number(fInfo.isDir());
						if (options->embedLinkedFiles() && fInfo.isFile()) {
							DEBUG_WIN << "Embed file";
							linkURL = folderName + copyFile(it->url().path(), folderPath, true);
						} else if (options->embedLinkedFolders() && fInfo.isDir()) {
							DEBUG_WIN << "Embed folder";
							linkURL = folderName + copyFile(it->url().path(), folderPath, true);
						} else {
							DEBUG_WIN << "Embed LINK";
							linkURL = it->url().prettyURL();
						}
					}
					break;
				case Item::Color:
//					stream << QString("<span style=\"color: %1;\">%2</span>").arg(it->color().name(), it->color().name());
					break;
				case Item::Unknown:
					stream << it->text(); // Since item text label already have formated HTML result
					break;
			}
			// Output content of file oriented items:
			if (it->type() == Item::Sound || it->type() == Item::File ||
			    it->type() == Item::Link  || it->type() == Item::Launcher) {
					if ( linkLook->showIcon() && ! linkIcon.isEmpty() ) {
						linkIcon = copyIcon(linkIcon, linkLook->iconSize(), folderPath);
						linkIcon = folderName + linkIcon;
						linkIcon = QString(
							"<img src=\"%1\" width=\"%2\" height=\"%2\" alt=\"\">")
								.arg(linkIcon, QString::number(linkLook->iconSize()));
						if (linkLook->onTopIcon())
							linkIcon += "<br>";
					}
					linkTitle = textToHTMLWithoutP(linkTitle);
					// Append address (useful for print version of the page/basket):
					if (options->formatForImpression() && it->type() == Item::Link &&
					     (!it->autoTitle() && it->title() != ItemFactory::titleForURL(it->url().prettyURL())) ) {
						// The address is on a new line, unless title is empty (empty lines was replaced by &nbsp;):
						if (linkTitle == "&nbsp;")
							linkTitle = "";
						else
							linkTitle = linkTitle + "<br>";
						linkTitle += "<i>" + it->url().prettyURL() + "</i>";
					}
					if (hAlign() == 2 && ! linkLook->onTopIcon())
						stream << QString("<a href=\"%1\">%2%3</a>")
						          .arg(linkURL, linkTitle == "" || linkTitle == " " ?
						                        QString("") : linkTitle, linkIcon);
					else
						stream << QString("<a href=\"%1\">%2%3</a>")
						          .arg(linkURL, linkIcon, linkTitle == "" || linkTitle == " " ?
						                                  QString("") : linkTitle);
			}
			// Output end tag of content:
//			if (showCheckBoxes())
//				stream << "</td>\n";
//			else
				stream << "</p>\n";
			// Output checkbox (if items are aligned at left):
/*			if (showCheckBoxes()) {
				if (hAlign() == 2)
					stream << QString(
						"        <td><input type=\"checkbox\"%1 disabled></td>\n")
							.arg(it->isChecked() ? " checked" : "");
				stream << "      </tr>\n";
			}*/
			if (it == lastShownItem())
				break;
		}

	// Ends the HTML file:
//	if (showCheckBoxes())
//		stream << "    </table>\n";
//	else
		stream << "    </div>\n";

	stream << QString(
		"    <p class=\"credits\">%1</p>\n"
		"  </body>\n"
		"</html>\n").arg(
			i18n("Made with %1, a KDE tool to take notes and keep a full range of data on hand.")
				.arg("<a href=\"http://basket.kde.org/\">%1</a> %2")
				.arg(kapp->aboutData()->programName(), VERSION));
	file.close();
}

QString Basket::copyIcon(const QString &iconName, int size, const QString &destFolder)
{
	if (iconName.isEmpty())
		return "";

	// Sometimes icon can be "favicons/www.kde.org", we replace the '/' by a '_'
	QString fileName = iconName; // QString::replace() isn't const, so I must copy the string before
	fileName = "ico" + QString::number(size) + "_" + fileName.replace("/", "_") + ".png";
	QString fullPath = destFolder + fileName;
	DesktopIcon(iconName, size).save(fullPath, "PNG"); // FIXME: I don't check if the file already exists!
	return fileName;
}

/** DONE: Sometimes we can call two times copyFile() with the same srcPath and destFolder
  *       (eg. when exporting basket to HTML with two links to same filename
  *            (but not necesary same path, as in "/home/foo.txt" and "/foo.txt") )
  *       The first copy isn't yet started, so the dest file isn't created and this method
  *       returns the same filename !!!!!!!!!!!!!!!!!!!!
  */

QString Basket::copyFile(const QString &srcPath, const QString &destFolder, bool createIt)
{
	QString fileName = Utilities::fileNameForNewFile(KURL(srcPath).fileName(), destFolder);
	QString fullPath = destFolder + fileName;

	if (createIt) {
		// We create the file to be sure another very near call to copyFile() willn't choose the same name:
		QFile file(KURL(fullPath).path());
		if ( file.open(IO_WriteOnly) )
			file.close();
		// And then we copy the file AND overwriting the file we juste created:
		new KIO::FileCopyJob(
			KURL(srcPath), KURL(fullPath), 0666, /*move=*/false,
			/*overwrite=*/true, /*resume=*/true, /*showProgress=*/true );
	} else
		/*KIO::CopyJob *copyJob = */KIO::copy(KURL(srcPath), KURL(fullPath)); // Do it as before

	return fileName;
}


/* Unfocus the previously focused item (unless it was null)
 * and focus the new @param item (unless it is null) if hasFocus()
 * Update m_focusedItem to the new one
 */
void Basket::setFocusedItem(Item *item) // void Basket::changeFocusTo(Item *item)
{
	if (item != 0L && !item->isShown()) // Don't focus a not visible item!
		return;
	if (m_focusedItem != 0L)
		m_focusedItem->setFocused(false);
	if (hasFocus() && item != 0L)
		item->setFocused(true);
	m_focusedItem = item;
}

/* If no shown item is currently focused, try to find a shown item and focus it
 * Also update m_focusedItem to the new one (or null if there isn't)
 */
void Basket::focusAnItem()
{
	if (countShown() == 0) {   // No item to focus
		setFocusedItem(0L);
		return;
	}

	if (m_focusedItem == 0L) { // No focused item yet : focus the first shown
		setFocusedItem(firstShownItem());
		return;
	}

	// Search a visible item to focus if the focused one isn't shown :
	Item *toFocus = m_focusedItem;
	if (!toFocus->isShown())
		toFocus = nextShownItemFrom(m_focusedItem, 1);
	if (toFocus == 0L)
		toFocus = nextShownItemFrom(m_focusedItem, -1);
	setFocusedItem(toFocus);
}

void Basket::editItem(Item *item, bool editAnnotations)
{
	if (isLocked())
		return;

	if (m_editor != 0L) // It arrive toolbar of another editor stay shown...
		closeEditor();

	setFocusedItem(item); // Focus and select item (in case of new inserted item)
	unselectAllBut(item);

	// If possible and if wanted, use inline editor :
	if ( true/*Settings::useInlineEditors()*/ && !editAnnotations && Global::mainContainer->isShown() &&
	     (item->type() == Item::Text || item->type() == Item::Html) ) {
		if (item->type() == Item::Text)
			m_editor = new ItemTextEditor(item, Global::mainContainer, viewport(), m_stackedKeyEvent);
		else
			m_editor = new ItemHtmlEditor(item, Global::mainContainer, viewport(), m_stackedKeyEvent);
		Global::mainContainer->addDockWindow(m_editor->toolbar());
/*		// DISABLE all SHOWN items:
		for (Item *it = firstShownItem(); it != 0L; it = it->next())
			if (it->isShown()) {
				it->setEnabled(false);
				if (it == lastShownItem())
					break;
			}*/
		// The editor :
		m_editor->editorWidget()->setPaletteBackgroundColor(item->isAlternate() ? altColor() : color());
		m_editor->editorWidget()->reparent( viewport(), QPoint(0,0), true );
		addChild(m_editor->editorWidget(), 0, 0);
		placeEditor();
//		((QFrame*)m_editor->editorWidget())->setFrameShape(QFrame::NoFrame);
//		((QFrame*)m_editor->editorWidget())->setLineWidth(1);
		// To avoid a one line text with an hozirontal scrollBar that take all the widget :
		if (item->type() == Item::Text || item->type() == Item::Html)
			((QScrollView*)m_editor->editorWidget())->setHScrollBarMode(QScrollView::AlwaysOff);
		m_editor->editorWidget()->show();
		m_editor->editorWidget()->raise();
		m_editor->goFirstFocus();
		connect( m_editor, SIGNAL(focusOut()), this, SLOT(closeEditor()) );
		connect( (QTextEdit*)(m_editor->editorWidget()), SIGNAL(textChanged()), this, SLOT(placeEditor()) );
//		if (m_stackedKeyEvent) {
//			kapp->postEvent(m_editor->editorWidget(), m_stackedKeyEvent);
		m_stackedKeyEvent = 0L;
//		}
		kapp->processEvents();   // Show the editor toolbar before ensuring the item is visible
		ensureVisibleItem(item); //  because toolbar can create a new line and then partially hide the item
		Global::mainContainer->updateStatusBarHint(); // Display the edition hint
	} else {
		ItemEditDialog *dialog = new ItemEditDialog(item, editAnnotations, item, m_stackedKeyEvent);
		dialog->exec();
		m_stackedKeyEvent = 0L;
		delete dialog;
	}
}




void Basket::openItem()
{
	for (Item *it = firstItem(); it != 0L; it = it->next())
		if (it->isSelected())
			it->slotOpen();
}

void Basket::openItemWith()
{
	if (m_countSelecteds != 1) // TODO: Do better for multiple open with...
		return;

	// Take THE selected item
	for (Item *it = firstItem(); it != 0L; it = it->next())
		if (it->isSelected()) {
			it->slotOpenWith();
			return;
		}
}

void Basket::saveItemAs()
{
	if (m_countSelecteds != 1) // TODO: Do better for multiple save as...
		return;

	// Take THE selected item
	for (Item *it = firstItem(); it != 0L; it = it->next())
		if (it->isSelected()) {
			it->slotSaveAs();
			return;
		}
}

void Basket::moveOnTop()
{
	if (m_countSelecteds == 0)
		return;

	Item *endOfBrowse = firstShownItem();
	Item *topItem     = firstItem();
	Item *prev;
	for (Item *it = lastShownItem(); it != 0L; ) {
		prev = it->previous();
		if (it->isSelected()) {
			m_insertAtItem = topItem;
			m_insertAfter  = false;
			changeItemPlace(it);
			topItem = it;
		}
		if (it == endOfBrowse)
			break;
		it = prev;
	}
	ensureVisibleItem(firstShownItem());
	ensureVisibleItem(m_focusedItem);
}

void Basket::moveOnBottom()
{
	if (m_countSelecteds == 0)
		return;

	Item *endOfBrowse = lastShownItem();
	Item *bottomItem  = lastItem();
	Item *next;
	for (Item *it = firstShownItem(); it != 0L; ) {
		next = it->next();
		if (it->isSelected()) {
			m_insertAtItem = bottomItem;
			m_insertAfter  = true;
			changeItemPlace(it);
			bottomItem = it;
		}
		if (it == endOfBrowse)
			break;
		it = next;
	}
	ensureVisibleItem(lastShownItem());
	ensureVisibleItem(m_focusedItem);
}

void Basket::moveItemUp()
{
	if (m_countSelecteds == 0)
		return;

	// Begin from the top (important move all selected items one item up
	// AND to quit early if a selected item is the first shown one
	for (Item *it = firstShownItem(); it != 0L; it = it->next()) {
		if (it->isSelected() && it->isShown()) { // it->isShown() not necessary, but in case...
			if (it == firstShownItem())
				return; // No way...
			m_insertAtItem = nextShownItemFrom(it, -1); // Previous shown item
			if (m_insertAtItem == 0L) { // Should not appends, since it's not the first shown item,
				resetInsertTo();        // there SHOULD be one before
				return;
			}
			m_insertAfter  = false;
			changeItemPlace(it);
		}
		if (it == lastShownItem())
			break;
	}
	ensureVisibleItem(m_focusedItem);
}

void Basket::moveItemDown()
{
	if (m_countSelecteds == 0)
		return;

	// Begin from the bottom (important move all selected items one item down
	// AND to quit early if a selected item is the last shown one
	for (Item *it = lastShownItem(); it != 0L; it = it->previous()) {
		if (it->isSelected() && it->isShown()) { // it->isShown() not necessary, but in case...
			if (it == lastShownItem())
				return; // No way...
			m_insertAtItem = nextShownItemFrom(it, 1); // Next shown item
			if (m_insertAtItem == 0L) { // Should not appends, since it's not the last shown item,
				resetInsertTo();        // there SHOULD be one before
				return;
			}
			m_insertAfter  = true;
			changeItemPlace(it);
		}
		if (it == firstShownItem())
			break;
	}
	ensureVisibleItem(m_focusedItem);
}


/*********** Work to re-load items when theire files have changed */
// TODO: class BasketUpdater
// FIXME: When a rename (de, new, del), if cutted by two timer intervals ????? Restart with 50 ms ? Yes.
// FIXME: If a timer can't be launched ? Is this happens often ?

// Created :         In :                  Deleted in :
// FileEvent(...)    slot*edFile           event
//   QString(...)

void Basket::dontCareOfCreation(const QString &path) /////////// FIXME: TODO: URGENT: notYetInserted() ???,
{
	m_dontCare.append(path);
	if (Global::debugWindow)
		*Global::debugWindow << "Watcher>Add a <i>don't care creation</i> of file : <font color=violet>" + path + "</font>";
}

void Basket::slotModifiedFile(const QString &path)
{
	m_updateQueue.append( new FileEvent(FileEvent::Modified, path) );
	if ( ! m_updateTimer.isActive() )
		m_updateTimer.start(c_updateTime, true); // 200 ms, only once
	if (Global::debugWindow)
		*Global::debugWindow << "Watcher>Modified : <font color=blue>" + path + "</font>";

}
void Basket::slotCreatedFile(const QString &path)
{
	bool dontCare = false;
	if (m_dontCare.contains(path)) { // Don't care of creation of files we know we have created ourself
		m_dontCare.remove(path);
		dontCare = true;
	} else {
		m_updateQueue.append( new FileEvent(FileEvent::Created, path) );
		if ( ! m_updateTimer.isActive() )
			m_updateTimer.start(c_updateTime, true); // 200 ms, only once
	}
	if (Global::debugWindow) {
		*Global::debugWindow << "Watcher>Created : <font color=green>" + path + "</font>";
		if (dontCare)
			*Global::debugWindow << "\tBut don't care (created by the application and we know this) : ignore it";
	}
}
void Basket::slotDeletedFile(const QString &path)
{
	m_updateQueue.append( new FileEvent(FileEvent::Deleted, path) );
	if ( ! m_updateTimer.isActive() )
		m_updateTimer.start(c_updateTime, true); // 200 ms, only once
	if (Global::debugWindow)
		*Global::debugWindow << "Watcher>Deleted : <font color=red>" + path + "</font>";
}

void Basket::slotUpdateItems()
{
	/* This functions is called after changes on the disk
	 *
	 */

//	m_dontCare.clear();

	if (Global::debugWindow)
		*Global::debugWindow << "Updater : Begin";

	/* First, browse the queue and keep only one instance of each files
	    (to reload/change only once), placed in lists per actions to do.
	    Note : Events was added in m_updateQueue by chronogical order */
	FileEvent   *event;
	QStringList  toCreate;
	QStringList  toModify;
	QStringList  toDelete;
	QStringList  toRename;
	QStringList  renameTo;
	while (m_updateQueue.count() > 0) {
		event = m_updateQueue.take(0);

		int     eventType =   event->event;
		QString eventPath(event->filePath);
		/* The .basket file isn't an item */
		if (eventPath.endsWith("/.basket")) {
			;                                             // Do nothing
		/* Item must be created ? */
		} else if (eventType == FileEvent::Created) {
			if (toCreate.findIndex(eventPath) == -1)      // Create only once (return first index or -1 if not)
				toCreate.append(eventPath);
		/* Item must be updated ? */
		} else if (eventType == FileEvent::Modified) {
			if ( toCreate.findIndex(eventPath) == -1 &&   // If created and then modified, do not add
			     toModify.findIndex(eventPath) == -1    ) // If modified several times,    do not add
				toModify.append(eventPath);
		/* Item must be deleted ? */
		} else if (eventType == FileEvent::Deleted) {
			// First try to see if it is a file to rename
			//  In this case, events are : [delete foo], [create bar], [delete foo]
			if (m_updateQueue.count() >= 2) { // If theire is at least two next events
				FileEvent *createEvt  = m_updateQueue.take(0);
				FileEvent *delete2Evt = m_updateQueue.take(0);
				if (createEvt->event     == FileEvent::Created &&
				    delete2Evt->event    == FileEvent::Deleted &&
				    delete2Evt->filePath == event->filePath       ) {
					if (toRename.findIndex(eventPath) == -1 ) { // Rename only once
						toRename.append(eventPath);
						renameTo.append(createEvt->filePath);
					}
					delete event;
					delete createEvt;
					delete delete2Evt;
					continue; // Continue on top of the while(){}
				} else { // It isn't a rename : re-place the events in the stack to parse them the next time
					m_updateQueue.prepend(delete2Evt); // In the inverse order !
					m_updateQueue.prepend(createEvt);
					// It's a delete : perform the corresponding action :
				}
			}
			if (toCreate.findIndex(eventPath) != -1)  // Created and then deleted
				toCreate.remove(eventPath);           //  => No need to create
			if (toModify.findIndex(eventPath) != -1)  // Modified and then deleted
				toModify.remove(eventPath);           //  => No need to modify
			if (toDelete.findIndex(eventPath) == -1 ) // Delete only once
				toDelete.append(eventPath);
		}
		delete event;

	}

	if (Global::debugWindow) {
		QValueList<QString>::iterator path;
		QValueList<QString>::iterator name2 = renameTo.begin();
		*Global::debugWindow << "    Create :";
		for (path = toCreate.begin(); path != toCreate.end(); ++path)
			*Global::debugWindow << "\t" + (*path);
		*Global::debugWindow << "    Modify :";
		for (path = toModify.begin(); path != toModify.end(); ++path)
			*Global::debugWindow << "\t" + (*path);
		*Global::debugWindow << "    Delete :";
		for (path = toDelete.begin(); path != toDelete.end(); ++path)
			*Global::debugWindow << "\t" + (*path);
		*Global::debugWindow << "    Rename :";
		for (path = toRename.begin(); path != toRename.end(); ++path) {
			*Global::debugWindow << "\t" + (*path) + " to " + (*name2);
			++name2;
		}
	}

	Item *item;
	QValueList<QString>::iterator path;
	QValueList<QString>::iterator name2 = renameTo.begin();
	for (path = toCreate.begin(); path != toCreate.end(); ++path) {
		// fileNameForFullPath() :
		QString fileName = (*path).mid(fullPath().length()); // fileName is never a mirror
		ItemFactory::loadFile(fileName, this);
	}
	for (path = toModify.begin(); path != toModify.end(); ++path)
		if ( (item = itemForFullPath(*path)) != 0 ) {
			item->loadContent();
			itemSizeChanged(item);
		}
	for (path = toDelete.begin(); path != toDelete.end(); ++path)
		if ( (item = itemForFullPath(*path)) != 0 )
			delItem(item);
	for (path = toRename.begin(); path != toRename.end(); ++path) {
		if ( (item = itemForFullPath(*path)) != 0 )
			item->fileNameChanged( (*name2).mid(fullPath().length()) ); // Same
		++name2;
	} // We view renamed files ^^ and save the change:
	if (!toRename.isEmpty())
		save();

	if (Global::debugWindow)
		*Global::debugWindow << "Updater : End";
}

/******************************************************************/

#endif // #if 0

#include "basket.moc"
