/***************************************************************************
 *   Copyright (C) 2003 by Sbastien Laot                                 *
 *   slaout@linux62.org                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ITEMFACTORY_H
#define ITEMFACTORY_H

#include <qevent.h>
#include <kurl.h>
#include <qstringlist.h>

class QString;
class QPixmap;
class QColor;

class Basket;
class Item;

enum ItemType::Id;

/** Factory class to create (new, drop, past) or load BasketIem, and eventuelly save them (?)
  * @author Sbastien Laot
  */
namespace ItemFactory
{
	/** Functions to create a new item from a content.
	  * Content, if any, is saved to file but the item is not insterted in the basket, and the basket is not saved.
	  * Return 0 if the item has not been successfully created.
	  * In some cases, the returned item can be a group containing several items or the first item of a chained list.
	  * The method Basket::TODO() can insert several grouped or chained items without problem.
	  */
	Item* createItemText(     const QString &text,     Basket *parent);
	Item* createItemHtml(     const QString &html,     Basket *parent);
	Item* createItemLink(     const KURL    &url,      Basket *parent);
	Item* createItemLink(     const KURL    &url,      const QString &title, Basket *parent);
	Item* createItemImage(    const QPixmap &image,    Basket *parent);
	Item* createItemColor(    const QColor  &color,    Basket *parent);
	Item* createItemFromText( const QString &content,  Basket *parent); // Find automatically the type from the text meaning  // TODO: Return Item::List?
	Item* createItemLauncher( const KURL    &url,      Basket *parent);
	Item* createItemLauncher( const QString &command,  const QString &name, const QString &icon, Basket *parent);
	Item* createItemUnknown(  QMimeSource *source,     Basket *parent);
	/** Functions to create derived items from a content */
	Item* createItemLinkOrLauncher( const KURL &url,   Basket *parent);
	Item* copyFileAndLoad(    const KURL &url,         Basket *parent);
	Item* moveFileAndLoad(    const KURL &url,         Basket *parent);
	Item* loadFile(           const QString &fileName, Basket *parent); /// << Determine the content of the file (the file SHOULD exists) and return a item of the good type.
	Item* loadFile(           const QString &fileName, ItemType::Id type, Basket *parent ); /// <<  Create an item of type @p type. The file is not obliged to exist.
	/** Functions to create a new item from a drop or past event */
	Item* dropItem(QMimeSource *source, Basket *parent,
	               bool fromDrop = false, QDropEvent::Action action = QDropEvent::Copy, Item *itemSource = 0);
	bool movingItemsInTheSameBasket(QMimeSource *source, Basket *parent, QDropEvent::Action action);
	Item* dropURLs(KURL::List urls, Basket *parent, QDropEvent::Action action, bool fromDrop);
	Item* decodeContent(QDataStream &stream, ItemType::Id type, Basket *parent); /// << Decode the @p stream to an item or return 0 if a general loadFile() is sufficient.
	void consumeContent(QDataStream &stream, ItemType::Id type); /// << Decode the @p stream to an item or return 0 if a general loadFile() is sufficient.
	/** Functions to create an item file but not load it in an item object */
	QString createItemLauncherFile(const QString &command, const QString &name, const QString &icon, Basket *parent);
	/** Other useful functions */
	ItemType::Id typeForURL(const KURL &url, Basket *parent);
	bool         maybeText(const KURL &url);
	bool         maybeHtml(const KURL &url);
	bool         maybeImageOrAnimation(const KURL &url);
	bool         maybeAnimation(const KURL &url);
	bool         maybeSound(const KURL &url);
	bool         maybeLauncher(const KURL &url);
	QString      fileNameForNewItem(Basket *parent, const QString &wantedName);
	QString      createFileForNewItem(Basket *parent, const QString &extension, const QString &wantedName = "");
	QString      titleForURL(const KURL &url);
	QString      iconForURL(const KURL &url);
	QString      iconForCommand(const QString &command);
	bool         isIconExist(const QString &icon);
	QStringList  textToURLList(const QString &text); // @Return { url1, title1, url2, title2, url3, title3... }
	/** Insert GUI menu */
	Item* createEmptyItem(  ItemType::Id type, Basket *parent ); // Insert empty if of type Item::Type
	Item* importKMenuLauncher(Basket *parent);
	Item* importIcon(Basket *parent);
	Item* importFileContent(Basket *parent);
}

#endif // ITEMFACTORY_H
